from ovary_analysis.measure.vol_measure import measure_dir

if __name__ == "__main__":
    FOLLICLE_SEGMENTATIONS_DIR = "/cluster/work/cobi/kevin/ivf/cross_validation_20220126/follicle_crossval_mask/fold_0/all_segmentations"

    ground_truth_df = measure_dir(
        dir_path=FOLLICLE_SEGMENTATIONS_DIR,
        file_pattern="*.h5",
        follicles_key="follicle_labels_rescaled"
    )
    ground_truth_df.to_csv("fold_0_all_measurements_ground_truth_20220404.csv")

    segmentation_df = measure_dir(
        dir_path=FOLLICLE_SEGMENTATIONS_DIR,
        file_pattern="*.h5",
        follicles_key="follicle_segmentation_rescaled"
    )
    segmentation_df.to_csv("fold_0_all_measurements_segmentation_20220404.csv")





from ovary_analysis.post_process.follicles import post_process_follicles_cross_val


CROSSVAL_DIR_PATH = '/cluster/work/cobi/kevin/ivf/cross_validation_20220126/follicle_crossval_mask'
FOLD_DIR_PATTERN = 'fold_*'
RAW_IM_DIR_NAME = 'test'
PREDICTIONS_DIR_NAME = 'predictions'
OUTPUT_DIR_NAME = 'segmentations'
PROB_THRESHOLD = 0.5
VOLUME_THRESHOLD = 30
PREDICTION_INDEX = 1
APPLY_SOFTMAX = True
RAW_IM_KEY = 'raw_rescaled'
FOLLICLES_LABELS_KEY = 'follicle_labels_rescaled'
OVARY_SEGMENTATION_KEY = 'ovary_labels_rescaled'
PREDICTIONS_KEY = 'predictions'

post_process_follicles_cross_val(
    crossval_dir_path=CROSSVAL_DIR_PATH,
    fold_dir_pattern=FOLD_DIR_PATTERN,
    raw_im_dir_name=RAW_IM_DIR_NAME,
    predictions_dir_name=PREDICTIONS_DIR_NAME,
    output_dir_name=OUTPUT_DIR_NAME,
    prob_threshold=PROB_THRESHOLD,
    volume_threshold=VOLUME_THRESHOLD,
    prediction_index=PREDICTION_INDEX,
    apply_softmax=APPLY_SOFTMAX,
    raw_im_key=RAW_IM_KEY,
    follicle_labels_key=FOLLICLES_LABELS_KEY,
    ovary_segmentation_key=OVARY_SEGMENTATION_KEY,
    prediction_key=PREDICTIONS_KEY
)



import os
from typing import Tuple

import numpy as np
import open3d as o3d
import pandas as pd
from scipy.spatial import ConvexHull
from skimage.segmentation import flood_fill
from tqdm.auto import tqdm

from ..io.hdf5_io import write_multi_dataset_hdf
from ..io.label_im import load_label_im
from ..utils.image import rescale_label_im
from ..utils.time import date_to_datestring


def generate_mesh(coords: np.ndarray) -> o3d.geometry.TriangleMesh:
    hull = ConvexHull(coords)
    vertices = hull.points[hull.vertices]

    vertex_lookup = {v: i for i, v in enumerate(hull.vertices)}

    faces = np.zeros_like(hull.simplices, dtype=np.int)

    for i, row in enumerate(hull.simplices):
        for j, element in enumerate(row):
            faces[i, j] = vertex_lookup[element]

    verts_o3d = o3d.utility.Vector3dVector(vertices)
    faces_o3d = o3d.utility.Vector3iVector(faces)

    mesh = o3d.geometry.TriangleMesh(vertices=verts_o3d, triangles=faces_o3d)

    return mesh


def mesh_to_binary(
    mesh: o3d.geometry.TriangleMesh, image_shape: Tuple[int, int, int]
) -> np.ndarray:
    assert mesh.is_watertight()

    voxel_grid = o3d.geometry.VoxelGrid.create_from_triangle_mesh(
        mesh, voxel_size=1
    )

    # query all voxels in the bounding box around the mesh
    min_bound = np.floor(voxel_grid.get_min_bound()) - 1
    max_bound = np.ceil(voxel_grid.get_max_bound()) + 1
    z_range = np.arange(min_bound[0], max_bound[0])
    x_range = np.arange(min_bound[1], max_bound[1])
    y_range = np.arange(min_bound[2], max_bound[2])

    all_coords = []
    for z in z_range:
        for x in x_range:
            for y in y_range:
                all_coords.append([z, x, y])

    coords_array = np.asarray(all_coords, dtype=np.int)
    in_mesh = voxel_grid.check_if_included(
        o3d.utility.Vector3dVector(coords_array)
    )
    coords_in_mesh = coords_array[in_mesh]

    # create the binary image
    binary_im = np.zeros(image_shape, dtype=np.int)
    binary_im[
        coords_in_mesh[:, 0], coords_in_mesh[:, 1], coords_in_mesh[:, 2]
    ] = 1

    # fill in the volume
    mesh_centroid = tuple(mesh.get_center().astype(np.int))
    filled_binary_im = flood_fill(binary_im, mesh_centroid, 1)

    return filled_binary_im


def fill_ovary_labels(sparse_ovary_labels: np.ndarray) -> np.ndarray:
    """Fill in sparse ovary labels to create a contiguous volume

    Parameters
    ----------
    sparse_ovary_labels :np.ndarray
        Label image where the outline of the ovary is sparsely labeled
        with non-zero values. All other voxels are 0.

    Returns
    -------
    filled_ovary_labels : np.ndarray
        Label image where all voxels corresponding to the ovary are labeled
    """
    coords = np.argwhere(sparse_ovary_labels)
    mesh = generate_mesh(coords)
    filled_ovary_labels = mesh_to_binary(
        mesh, image_shape=sparse_ovary_labels.shape
    )
    return filled_ovary_labels


def convert_labels(
    row: pd.Series,
    target_px_size: Tuple[float, float, float],
    output_dir: os.PathLike,
) -> os.PathLike:
    """Convert the raw labels to the training format.

    Parameters
    ----------
    row : pd.Series
        The row in the dataset table corresponding to the label image to be converted.
        The row should contain the following columns:
        ['patient', 'cycle', 'side', 'day', 'path_labels']
    target_px_size : Tuple[float, float, float]
        The target voxel size in mm/px. Should be in ZYX order.
    output_dir : os.PathLike
        The path to save the converted file to.

    Returns
    -------
    output_file_path : os.PathLike
        The path the converted file was saved to. The file follows the pattern:
        f'output_dir/{patient_id}_{cycle}_{side}_{date_string}_converted_labels.h5'
    """
    patient_id = row['patient']
    cycle = row['cycle']
    side = row['side']
    day = row['day']
    raw_labels_fpath = row['path_labels']

    label_im = load_label_im(raw_labels_fpath)

    # get the follicle labels
    follicle_labels = label_im[1, :, :, :].astype(np.uint8)
    follicle_labels[follicle_labels > 0] = 1

    # get the ovary labels (saved as contour)
    sparse_ovary_labels = label_im[0, :, :, :]

    # fill in the ovary labels
    filled_ovary_labels = fill_ovary_labels(sparse_ovary_labels).astype(
        np.uint8
    )

    # create the filename
    date_string = date_to_datestring(day)
    output_fname = (
        f'{patient_id}_{cycle}_{side}_{date_string}_converted_labels.h5'
    )

    # rescale the label images
    im_px_size = (row['px_size_z'], row['px_size_x'], row['px_size_y'])
    rescaled_follicle_labels = rescale_label_im(
        follicle_labels, im_px_size=im_px_size, target_px_size=target_px_size
    ).astype(bool)
    rescaled_ovary_labels = rescale_label_im(
        filled_ovary_labels,
        im_px_size=im_px_size,
        target_px_size=target_px_size,
    ).astype(bool)

    # create the dataset dicts for the rescaled labels
    rescaled_follicle_labels_data = {
        'data': rescaled_follicle_labels,
        'attrs': {'px_size': target_px_size},
    }
    rescaled_ovary_labels_data = {
        'data': rescaled_ovary_labels,
        'attrs': {'px_size': target_px_size},
    }

    # save the converted labels
    output_file_path = os.path.join(output_dir, output_fname)
    write_multi_dataset_hdf(
        file_path=output_file_path,
        compression='gzip',
        raw_labels=label_im,
        follicle_labels=follicle_labels,
        ovary_labels=filled_ovary_labels,
        rescaled_follicle_labels=rescaled_follicle_labels_data,
        rescaled_ovary_labels=rescaled_ovary_labels_data,
    )

    return output_file_path


def batch_convert_labels(
    dataset_table: pd.DataFrame,
    target_px_size: Tuple[float, float, float],
    output_dir: os.PathLike,
) -> pd.DataFrame:
    """Convert a batch of label images

    Parameters
    ----------
    dataset_table : pd.DataFrame
        The table of label images to convert. Must contain columns:
        'patient', 'cycle', 'side', 'day', 'path_labels'
    target_px_size : Tuple[float, float, float]
        The target voxel size in mm/px. Should be in ZYX order.
    output_dir : os.PathLike
        The path to the directory to save the converted files to.

    Returns
    -------
    dataset_table : pd.DataFrame
        The input dataset_table with a `converted_labels_path` added
        containing the path to the converted files.
    """

    file_paths = []
    for i, row in tqdm(dataset_table.iterrows(), total=len(dataset_table)):

        file_path = convert_labels(
            row=row,
            target_px_size=target_px_size,
            output_dir=output_dir,
        )
        file_paths.append(file_path)
    dataset_table['converted_labels_path'] = file_paths

    return dataset_table



from typing import Any, List, Tuple

import numpy as np
from skimage.io import imread
from tifffile import TiffFile


def load_raw_im(raw_im_paths: List[str]) -> np.ndarray:
    """Load a raw image into a numpy array

    Parameters
    ----------
    raw_im_paths : List[str]
        A list of the paths to the z slices for the raw image.
        These should be ordered in ascending z order.

    Returns
    -------
    im_array : np.ndarray
        The image as a numpy array. Ordered ZXY.
    """
    # load the files
    ims = []
    for f in raw_im_paths:
        ims.append(imread(f))
    im_array = np.asarray(ims)
    return im_array


def parse_tiff_metadata(
    fpath: str, tag_name: Any = 270, separator: str = '\r\n'
) -> Tuple[float, float, float, int]:
    # get the metadata
    with TiffFile(fpath) as tif:
        tags = tif.pages[0].tags
        raw_metadata = tags[tag_name].value
    metadata = raw_metadata.split(separator)

    # get the pixel sizes
    px_size_x = _get_metadata_field(metadata, 'Pixel length - X [mm]', float)
    px_size_y = _get_metadata_field(metadata, 'Pixel length - Y [mm]', float)
    px_size_z = _get_metadata_field(metadata, 'Pixel length - Z [mm]', float)

    # get the number of slices
    n_z = _get_metadata_field(metadata, 'Number of Timepoints', int)

    return px_size_x, px_size_y, px_size_z, n_z


def _get_metadata_field(
    metadata: List[str], field_name: str, field_type
) -> Any:
    for field in metadata:
        if field.startswith(field_name):
            split_field = field.split(' ')
            field_value = field_type(split_field[-1])
            break
    return field_value



import glob
import os
from typing import Tuple
from tqdm import tqdm

import pandas as pd

from ..io.raw_im import parse_tiff_metadata
from .file_names import (
    calculate_day_offsets,
    is_cycle_folder,
    parse_cycle_folder,
    parse_day_dirs,
    parse_label_file_name,
    parse_z_slice,
    validate_labels_file,
    validate_z_slices,
)


def create_raw_file_table(
    root_dir: str, make_file_table: bool = True
) -> Tuple[
    pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame
]:
    patient_folders = [f.path for f in os.scandir(root_dir) if f.is_dir()]

    day_table_data = []
    file_table_data = []
    bad_cycle_folders = []
    bad_day_folders = []
    bad_z_slices = []

    for patient in tqdm(patient_folders):
        cycle_folders = [
            f.path
            for f in os.scandir(patient)
            if (f.is_dir() and is_cycle_folder(f.path))
        ]
        patient_id = os.path.basename(patient)

        for cycle in tqdm(cycle_folders, leave=False):
            if is_cycle_folder(cycle):
                day_dirs = [f.path for f in os.scandir(cycle) if f.is_dir()]
                cycle_idx, side = parse_cycle_folder(cycle)

                dates, good_day_dirs, bad_day_dirs = parse_day_dirs(day_dirs)
                date_offsets = calculate_day_offsets(dates)

                # save the paths for the day folders that couldn't be parsed
                for bad_date in bad_day_dirs:
                    bad_day_folders.append(
                        {
                            'patient': patient_id,
                            'cycle': cycle_idx,
                            'side': side,
                            'folder_path': bad_date,
                        }
                    )

                for day_path, day_abs, day_rel in tqdm(
                    zip(good_day_dirs, dates, date_offsets), leave=False
                ):
                    image_file_pattern = os.path.join(day_path, '*.tif*')
                    image_files = glob.glob(image_file_pattern)

                    z_slices = []
                    metadata_parsed = False
                    for idx, im_slice in enumerate(image_files):
                        success, z_idx = parse_z_slice(im_slice)
                        if success is True:
                            z_slices.append(z_idx)

                            if metadata_parsed is False:
                                (
                                    px_size_x,
                                    px_size_y,
                                    px_size_z,
                                    n_z,
                                ) = parse_tiff_metadata(im_slice)
                                metadata_parsed = True
                        else:
                            bad_z_slices.append(
                                {
                                    'patient': patient_id,
                                    'cycle': cycle_idx,
                                    'side': side,
                                    'day': day_abs,
                                    'day_rel': day_rel,
                                    'path': day_path,
                                    'file_path': im_slice,
                                }
                            )

                    im_files_valid = validate_z_slices(z_slices)

                    if metadata_parsed is False:
                        px_size_x = 0
                        px_size_y = 0
                        px_size_z = 0
                        n_z = 0

                    day_table_data.append(
                        {
                            'patient': patient_id,
                            'cycle': cycle_idx,
                            'side': side,
                            'day': day_abs,
                            'day_rel': day_rel,
                            'path': day_path,
                            'px_size_x': px_size_x,
                            'px_size_y': px_size_y,
                            'px_size_z': px_size_z,
                            'n_z': n_z,
                            'n_z_files': len(z_slices),
                            'z_slices_valid': im_files_valid,
                        }
                    )

                    if make_file_table:
                        for z, im in zip(z_slices, image_files):
                            file_table_data.append(
                                {
                                    'patient': patient_id,
                                    'cycle': cycle_idx,
                                    'side': side,
                                    'day': day_abs,
                                    'day_rel': day_rel,
                                    'z': z,
                                    'dir_path': day_path,
                                    'im_path': im,
                                }
                            )
            else:
                bad_cycle_folders.append(
                    {'patient': patient_id, 'folder_path': cycle}
                )
    day_table = pd.DataFrame.from_records(day_table_data)
    bad_cycle_table = pd.DataFrame.from_records(bad_cycle_folders)
    bad_day_table = pd.DataFrame.from_records(bad_day_folders)
    bad_z_table = pd.DataFrame.from_records(bad_z_slices)
    if make_file_table:
        file_table = pd.DataFrame.from_records(file_table_data)
    else:
        file_table = None

    return day_table, file_table, bad_cycle_table, bad_day_table, bad_z_table


def create_labels_file_table(root_dir: str) -> pd.DataFrame:
    patient_folders = [f.path for f in os.scandir(root_dir) if f.is_dir()]
    day_table_data = []

    for patient in patient_folders:
        cycle_folders = [
            f.path
            for f in os.scandir(patient)
            if (f.is_dir() and is_cycle_folder(f.path))
        ]
        patient_id = os.path.basename(patient)

        for cycle in cycle_folders:
            cycle_idx, side = parse_cycle_folder(cycle)

            image_file_pattern = os.path.join(cycle, '*.tif*')
            image_files = glob.glob(image_file_pattern)

            for im in image_files:
                day_abs = parse_label_file_name(im)
                file_valid, n_z = validate_labels_file(im)

                day_table_data.append(
                    {
                        'patient': patient_id,
                        'cycle': cycle_idx,
                        'side': side,
                        'day': day_abs,
                        'path': os.path.abspath(im),
                        'n_z': n_z,
                        'file_valid': file_valid,
                    }
                )
    day_table = pd.DataFrame.from_records(day_table_data)
    return day_table



from ovary_analysis.train.follicle_seg import create_cross_validation_datasets

training_job_settings = {
    'n_cpus': 4,
    'mem_per_cpu': 6144,
    'n_gpus': 1,
    'job_time': [96, 0]
}

prediction_job_settings = {
    'n_cpus': 10,
    'mem_per_cpu': 6144,
    'n_gpus': 8,
    'job_time': [1, 30]
}

CROSSVAL_DIR = '/cluster/work/cobi/kevin/ivf/cross_validation_20211031/ovary_crossval'
OVARY_DATASETS_DIR = '/cluster/work/cobi/kevin/ivf/cross_validation_20211031/converted_data/ovary_datasets_20211101'
TRAINING_CONFIG_PATH = '/cluster/work/cobi/kevin/ivf/cross_validation_20211031/follicle_crossval_20211118/train_follicle_crossval_20211118.yaml'
PREDICTION_CONFIG_PATH = '/cluster/work/cobi/kevin/ivf/cross_validation_20211031/follicle_crossval_20211118/predict_follicle_crossval_20211118.yaml'
OUTPUT_BASE_DIR = '/cluster/work/cobi/kevin/ivf/cross_validation_20211031/follicle_crossval_20211118'

TRAINING_RUNNER = '/cluster/home/kyamauch/.local/lib/python3.8/site-packages/pytorch3dunet/train.py'
PREDICTION_RUNNER = '/cluster/home/kyamauch/.local/lib/python3.8/site-packages/pytorch3dunet/predict.py'

create_cross_validation_datasets(
    crossval_dir=CROSSVAL_DIR,
    ovary_data_table_name='ovary_crossvalidation_20211101.csv',
    follicle_data_table_name='follicle_cross_validation_20211118.csv',
    ovary_seg_dir_name='segmentations',
    ovary_datasets_dir=OVARY_DATASETS_DIR,
    training_config_path=TRAINING_CONFIG_PATH,
    training_runner=TRAINING_RUNNER,
    training_job_settings=training_job_settings,
    prediction_config_path=PREDICTION_CONFIG_PATH,
    prediction_runner=PREDICTION_RUNNER,
    prediction_job_settings=prediction_job_settings,
    output_base_dir=OUTPUT_BASE_DIR,
    raw_name='raw_rescaled',
    follicle_name='follicle_labels_rescaled',
    ovary_name='ovary_labels_rescaled',
    ovary_seg_name='ovary_segmentation_rescaled',
    weights_name='weights',
    mask_raw_im=False,
    min_shape=[80, 80, 80],
    follicle_erosion=2,
    follicle_dilation=8,
    ovary_weight=0.2,
    follicle_weight=0.2,
    boundary_weight=1,
)


import numpy as np
import pymeshfix
from skimage.measure import marching_cubes
from trimesh import Trimesh
from trimesh.smoothing import filter_taubin


def binary_mask_to_surface(
    object_mask: np.ndarray, n_mesh_smoothing_interations: int = 50
) -> Trimesh:
    """Convert surface of a 3D binary mask (segmented object) into a watertight mesh.

    Parameters
    ----------
    object_mask  : BinaryMask
        A 3D binary image corresponding to the object you want to mesh.
    n_mesh_smoothing_interations : int
        The number of interations of smooting to perform. Smoothing is
        done by the trimesh taubin filter:
        https://trimsh.org/trimesh.smoothing.html#trimesh.smoothing.filter_taubin
        Default value is 50.

    Returns
    -------
    mesh : trimesh.Trimesh
        The resulting mesh as a trimesh.Trimesh object.
        https://trimsh.org/trimesh.base.html#github-com-mikedh-trimesh
    """
    vertices, faces, _, _ = marching_cubes(object_mask, 0)

    vertices_clean, faces_clean = pymeshfix.clean_from_arrays(vertices, faces)

    # create the mesh object
    mesh = Trimesh(vertices=vertices_clean, faces=faces_clean)

    # optionally clean up the mesh
    if n_mesh_smoothing_interations > 0:
        filter_taubin(mesh, iterations=n_mesh_smoothing_interations)

    return mesh



import os
from typing import Union, Dict, Any, List
import warnings

import h5py
import numpy as np
import pandas as pd
from skimage.morphology import binary_dilation, cube, binary_erosion
from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from .training_utils import copy_datasets, split_table_into_k_folds
from ..io.hdf5_io import load_raw_im, write_multi_dataset_hdf
from ..post_process.follicles import remove_small_follicles
from ..utils.config_utils import (
    _load_and_update_training_config,
    _load_and_update_prediction_config,
)
from ..utils.image import pad_label_im, pad_im_to_min_size
from ..utils.submission_utils import (
    _write_predict_script,
    _write_train_script,
    _write_submission_script,
)
from ..utils.time import date_to_datestring


def create_ovary_training_dataset(
    row: pd.Series,
    raw_im_dir: Union[str, os.PathLike],
    label_im_dir: Union[str, os.PathLike],
    output_dir: Union[str, os.PathLike],
    raw_im_name: str = 'raw',
    min_follicle_volume: int = 0,
    follicle_labels_name: str = 'rescaled_follicle_labels',
    ovary_labels_name: str = 'rescaled_ovary_labels',
    min_im_size: List[int] = [80, 80, 80],
    padding_weight: float = 0,
    muscle_weight: float = 0.9,
    ovary_weight: float = 0.9,
    follicle_weight: float = 1,
    boundary_weight: float = 1,
) -> str:
    """Create a training/validation dataset for the ovary segmentation

    Parameters
    ----------
    row : pd.Series
        The row in the file table to create the ovary dataset from.
    raw_im_dir : Union[str, os.PathLike]
        The path to the directory containing the converted raw images.
    label_im_dir : Union[str, os.PathLike]
        The path to the directory containing the converted label images
    output_dir : Union[str, os.PathLike]
        The path to the directory to save the ovary datasets to.
    raw_im_name : str
        The dataset key in the raw image file containig the raw image.
        The rescaled raw image will be accessed from raw_im_name + 'rescaled'
    min_follicle_volume : int
        The minimum number of voxels a follicle needs to have in the label image
        to pass QC. All follicles <= min_follicle_volume are remove. The default
        value is 0.
    follicle_labels_name : str
        The dataset key in the label image file for the rescaled follicle labels.
        The default value is 'rescaled_follicle_labels'.
    ovary_labels_name : str
        The dataset key in the label image file for the rescaled ovary labels.
        The default value is 'rescaled_ovary_labels'.
    min_im_size : List[int]
        The minimum shape for images. This ensures you will be able ot create
        at least one tile. Images with smaller shape are padded with zeros.

    Returns
    -------
    dataset_path : str
        The path to the dataset that was created.
    """
    patient_id = row['patient']
    cycle_id = row['cycle']
    side = row['side']
    date_string = date_to_datestring(row['day'])

    # load the image
    raw_im_fname = os.path.basename(row['converted_im_path'])
    raw_im_fpath = os.path.join(raw_im_dir, raw_im_fname)
    raw_im, raw_attrs, rescaled_im, rescaled_attrs = load_raw_im(
        raw_im_fpath, raw_name=raw_im_name
    )

    # load the labels
    labels_fname = os.path.basename(row['converted_labels_path'])
    labels_fpath = os.path.join(label_im_dir, labels_fname)
    with h5py.File(labels_fpath, 'r') as f_labels:
        follicle_labels = f_labels[follicle_labels_name][:]
        ovary_labels = f_labels[ovary_labels_name][:]

    # we have some label images that are missing a z slice
    # so we have to fix this as a special case
    follicle_labels = pad_label_im(rescaled_im, follicle_labels)
    ovary_labels = pad_label_im(rescaled_im, ovary_labels)

    # apply padding to make sure we meet our minimum tile size
    rescaled_im = pad_im_to_min_size(rescaled_im, min_size=min_im_size)
    ovary_labels = pad_im_to_min_size(ovary_labels, min_size=min_im_size)
    follicle_labels = pad_im_to_min_size(follicle_labels, min_size=min_im_size)

    follicle_labels[follicle_labels > 0] = 1
    follicle_labels_clean = remove_small_follicles(
        follicle_labels, volume_threshold=min_follicle_volume
    ).astype(np.uint8)

    # padding mask
    padding_mask_rescaled = np.zeros_like(rescaled_im, dtype=bool)
    padding_mask_rescaled[rescaled_im == 0] = True

    # create the weights image
    weights = create_ovary_weights(
        padding_im=padding_mask_rescaled,
        ovary_im=ovary_labels,
        follicles_im=follicle_labels_clean,
        padding_weight=padding_weight,
        muscle_weight=muscle_weight,
        ovary_weight=ovary_weight,
        follicle_weight=follicle_weight,
        boundary_weight=boundary_weight,
    )

    fname = f'{patient_id}_{cycle_id}_{side}_{date_string}_ovary.h5'
    file_path = os.path.join(output_dir, fname)

    raw_dataset = {'data': raw_im, 'attrs': raw_attrs}
    rescaled_dataset = {'data': rescaled_im, 'attrs': rescaled_attrs}

    write_multi_dataset_hdf(
        file_path=file_path,
        compression='gzip',
        raw=raw_dataset,
        raw_rescaled=rescaled_dataset,
        follicle_labels_rescaled=follicle_labels_clean,
        ovary_labels_rescaled=ovary_labels,
        weights_rescaled=weights,
    )

    return file_path


def batch_convert_ovary_datasets(
    dataset_table: pd.DataFrame,
    raw_im_dir: Union[str, os.PathLike],
    label_im_dir: Union[str, os.PathLike],
    output_dir: Union[str, os.PathLike],
    raw_im_name: str = 'raw',
    min_follicle_volume: int = 0,
    min_im_size: List[int] = [80, 80, 80],
    padding_weight: float = 0,
    muscle_weight: float = 0.9,
    ovary_weight: float = 0.9,
    follicle_weight: float = 1,
    boundary_weight: float = 1,
    follicle_labels_name: str = 'rescaled_follicle_labels',
    ovary_labels_name: str = 'rescaled_ovary_labels',
    dataset_path_name: str = 'dataset_path',
) -> pd.DataFrame:

    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    file_paths = []
    for i, row in tqdm(dataset_table.iterrows(), total=len(dataset_table)):

        file_path = create_ovary_training_dataset(
            row=row,
            raw_im_dir=raw_im_dir,
            label_im_dir=label_im_dir,
            output_dir=output_dir,
            raw_im_name=raw_im_name,
            min_follicle_volume=min_follicle_volume,
            follicle_labels_name=follicle_labels_name,
            ovary_labels_name=ovary_labels_name,
            min_im_size=min_im_size,
            padding_weight=padding_weight,
            muscle_weight=muscle_weight,
            ovary_weight=ovary_weight,
            follicle_weight=follicle_weight,
            boundary_weight=boundary_weight,
        )
        file_paths.append(file_path)
    dataset_table[dataset_path_name] = file_paths

    return dataset_table


def split_and_create_ovary_dataset(
    dataset_table: pd.DataFrame,
    raw_im_dir: str,
    output_base_dir: str,
    train_frac: float = 0.8,
    val_frac: float = 0.1,
    test_frac: float = 0.1,
) -> pd.DataFrame:
    assert (train_frac + val_frac + test_frac) == 1

    split_1 = train_frac
    split_2 = train_frac + val_frac

    # split the dataset
    train, validate, test = np.split(
        dataset_table.sample(frac=1, random_state=0),
        [int(split_1 * len(dataset_table)), int(split_2 * len(dataset_table))],
    )

    # create the training data
    train_dir = os.path.join(output_base_dir, 'train')
    os.mkdir(train_dir)
    train_output = batch_convert_ovary_datasets(
        dataset_table=train,
        raw_im_dir=raw_im_dir,
        output_dir=train_dir,
    )
    train_output['group'] = 'train'

    # create the validation data
    val_dir = os.path.join(output_base_dir, 'val')
    os.mkdir(val_dir)
    val_output = batch_convert_ovary_datasets(
        dataset_table=validate,
        raw_im_dir=raw_im_dir,
        output_dir=val_dir,
    )
    val_output['group'] = 'val'

    # create the test data
    test_dir = os.path.join(output_base_dir, 'test')
    os.mkdir(test_dir)
    test_output = batch_convert_ovary_datasets(
        dataset_table=test, raw_im_dir=raw_im_dir, output_dir=test_dir
    )
    test_output['group'] = 'test'

    output_df = pd.concat([train_output, val_output, test_output])

    return output_df


def create_ovary_weights(
    padding_im: np.ndarray,
    ovary_im: np.ndarray,
    follicles_im: np.ndarray,
    padding_weight: float = 0,
    muscle_weight: float = 0.9,
    ovary_weight: float = 0.9,
    follicle_weight: float = 1,
    boundary_weight: float = 1,
):
    padding_mask = padding_im.astype(np.bool)
    ovary_mask = ovary_im.astype(np.bool)
    follicles_mask = follicles_im.astype(np.bool)

    weights = np.zeros_like(ovary_mask, dtype=np.float)

    # apply the padding weights
    weights[padding_mask] = padding_weight

    # apply the muscle weights
    weights[np.logical_not(padding_mask)] = muscle_weight

    # apply the ovary weights
    weights[ovary_mask] = ovary_weight

    # apply the follicle weights
    weights[follicles_mask] = follicle_weight

    # apply the follicle boundary weights
    dilated_ovary = binary_dilation(ovary_mask, selem=cube(6))
    eroded_ovary = binary_erosion(ovary_mask, selem=cube(2))
    boundary_mask = np.logical_xor(dilated_ovary, eroded_ovary)
    weights[boundary_mask] = boundary_weight

    return weights


def create_cross_validation(
    data_table_path: Union[str, os.PathLike],
    dataset_dir: Union[str, os.PathLike],
    training_config_path: Union[str, os.PathLike],
    training_runner: str,
    training_job_settings: Dict[str, Any],
    prediction_config_path: Union[str, os.PathLike],
    prediction_runner: str,
    prediction_job_settings: Dict[str, Any],
    output_dir: Union[str, os.PathLike],
    df_output_fname: Union[str, os.PathLike],
    n_folds: int = 10,
    frac_train: float = 0.89,
    random_seed: int = 0,
):
    """Create the folders for a ovary cross validation experiment

    Parameters
    ----------
    data_table_path : Union[str, os.PathLike]
        The path to the dataset table to split.
    dataset_dir : Union[str, os.PathLike]
        The path to the folder containing the datasets to split
    training_config_path : Union[str, os.PathLike]
        The path to the training config that will be updated and stored in the
    training_runner : str
        The command to call the training runner. This us generally a bash
        or python script.
    training_job_settings : Dict[str, Any]
        The settings for each job that will be submitted for training the model.
        Must have the following keys: n_cpus, mem_per_cpu, n_gpus job_time
    prediction_config_path : Union[str, os.PathLike]
        The path to the training config that will be updated and stored in the
    prediction_runner : str
        The command to call the predict runner. This us generally a bash
        or python script.
    prediction_job_settings : Dict[str, Any]
        The settings for each job that will be submitted for predictions.
        Must have the following keys: n_cpus, mem_per_cpu, n_gpus job_time
    output_dir : Union[str, os.PathLike]
        The path to the directory to save the cross validation datasets to
    df_output_fname : Union[str, os.PathLike]
        The name of the final dataset table. This will be saved in the output_dir.
    n_folds : int
        The number of folds to split the data into. This will
    frac_train : float
        The fraction of the training portion of the fold to use for training. The
        rest is used for validation.
    random_seed : float
        The value for the random seed. The random number generator is used for
        splitting the data.
    """

    # load the dataframe
    df = pd.read_csv(data_table_path)
    n_datasets = len(df)
    print(f'table loaded, {n_datasets} found')

    # split the data into k partitions
    split_df = split_table_into_k_folds(df, n_folds, random_seed)

    training_scripts = []
    prediction_scripts = []
    for i in range(n_folds):
        print(f'starting fold {i}')
        train_val = split_df.copy()

        # get the test dataset
        test = split_df[i]
        del train_val[i]

        # split the train_val set
        train_val_df = pd.concat(train_val)
        frac_val = 1 - frac_train
        train, val = train_test_split(
            train_val_df, test_size=frac_val, random_state=random_seed
        )

        with warnings.catch_warnings():
            # set the group in the dataframes. We ignore the
            # SettingWithCopyWarning, as we do not intend
            # to propagate the change to the original df
            warnings.simplefilter("ignore")
            test['group'] = 'test'
            train['group'] = 'train'
            val['group'] = 'val'

        # make the folder for the group
        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)
        fold_dir_path = os.path.join(output_dir, f'fold_{i}')
        if not os.path.isdir(fold_dir_path):
            os.mkdir(fold_dir_path)

        # copy the test data
        test_data_path = os.path.join(fold_dir_path, 'test')
        copy_datasets(test, dataset_dir, test_data_path)

        # copy the train data
        train_data_path = os.path.join(fold_dir_path, 'train')
        copy_datasets(train, dataset_dir, train_data_path)

        # copy the val data
        val_data_path = os.path.join(fold_dir_path, 'val')
        copy_datasets(val, dataset_dir, val_data_path)

        print(f'fold {i} data copied')
        # recombine the dataframes and save
        fold_df = pd.concat([test, val, train])
        fold_df.to_csv(os.path.join(fold_dir_path, df_output_fname))

        # setup and save the training config
        checkpoint_dir = os.path.join(fold_dir_path, 'checkpoints')
        updated_training_config_path = _load_and_update_training_config(
            config_path=training_config_path,
            fold_index=i,
            checkpoint_dir=checkpoint_dir,
            train_file_dir=[train_data_path],
            val_file_dir=[val_data_path],
            config_output_directory=fold_dir_path,
        )
        training_script_path = _write_train_script(
            train_runner=training_runner,
            config_path=updated_training_config_path,
            output_dir=fold_dir_path,
        )
        training_scripts.append(training_script_path)

        # setup and save the prediction config
        model_path = os.path.join(checkpoint_dir, 'best_checkpoint.pytorch')
        prediction_output_path = os.path.join(fold_dir_path, 'predictions')
        if not os.path.isdir(prediction_output_path):
            os.mkdir(prediction_output_path)
        updated_prediction_config_path = _load_and_update_prediction_config(
            config_path=prediction_config_path,
            fold_index=i,
            model_path=model_path,
            image_files=[train_data_path, val_data_path, test_data_path],
            prediction_output_directory=prediction_output_path,
            config_output_directory=fold_dir_path,
        )
        prediction_script_path = _write_predict_script(
            predict_runner=prediction_runner,
            config_path=updated_prediction_config_path,
            output_dir=fold_dir_path,
        )
        prediction_scripts.append(prediction_script_path)

        print(f'fold{i} complete')

    print('writing submission scripts')
    training_submission_path = os.path.join(
        output_dir, 'training_submission_script.sh'
    )
    training_job_settings.update(
        {
            'runner_paths': training_scripts,
            'job_prefix': 'training',
            'script_path': training_submission_path,
        }
    )
    _write_submission_script(**training_job_settings)

    prediction_submission_path = os.path.join(
        output_dir, 'prediction_submission_script.sh'
    )
    prediction_job_settings.update(
        {
            'runner_paths': prediction_scripts,
            'job_prefix': 'prediction',
            'script_path': prediction_submission_path,
        }
    )
    _write_submission_script(**prediction_job_settings)

    print('done')



import multiprocessing as mp
from functools import partial

import pandas as pd

from ovary_analysis.convert.raw_data import convert_raw_dataset


if __name__ == "__main__":
    IMAGE_TABLE_PATH = '/nas/groups/iber/Projects/IVF/Bicycle_Study/Data/ultrasound_images_h5/image_table_raw_files_20220228.csv'
    OUTPUT_DIRECTORY = '/local0/kevin/ivf/full_data_converted_20220301/converted'
    TARGET_PX_SIZE = (0.157288, 0.157288, 0.157288)
    IM_PATH_KEY = "path"
    N_Z_KEY = "n_z_files"

    # load the image table
    image_table = pd.read_csv(IMAGE_TABLE_PATH)
    image_table_rows = [row for _, row in image_table.iterrows()]

    conversion_function = partial(
        convert_raw_dataset,
        target_px_size=TARGET_PX_SIZE,
        output_dir=OUTPUT_DIRECTORY,
        im_path_key=IM_PATH_KEY,
        n_z_key=N_Z_KEY
    )
    print(f"converting {len(image_table)} images")
    with mp.get_context('spawn').Pool() as pool:
        records = pool.map(conversion_function, image_table_rows)

    # records = []
    # for i, row in enumerate(image_table_rows):
    #     print(i)
    #     records.append(conversion_function(row))



from functools import partial
import glob
import multiprocessing as mp
import os
from typing import Tuple

import h5py

from ovary_analysis.io.hdf5_io import write_hdf
from ovary_analysis.utils.image import crop_raw_image


def create_follicle_dataset(
    raw_fpath: str,
    segmentation_dir: str,
    output_dir: str,
    min_shape: Tuple[int, int, int],
    mask_raw_im: bool = True
):
    # load the raw image
    with h5py.File(raw_fpath, "r") as f:
        raw_im = f["raw_rescaled"][:]

    # load the ovary segmentation
    raw_fname = os.path.basename(raw_fpath)
    ovary_segmentation_name = raw_fname.replace(".h5", "_segmentation.h5")
    ovary_segmentation_path = os.path.join(segmentation_dir, ovary_segmentation_name)
    with h5py.File(ovary_segmentation_path, "r") as f:
        ovary_segmentation = f["ovary_segmentation_rescaled"][:]

    ovary_mask = ovary_segmentation > 0
    cropped_raw = crop_raw_image(
        raw_im, ovary_mask, min_shape=min_shape, mask_periphery=mask_raw_im
    )

    output_fname = raw_fname.replace("_raw.h5", "_follicles.h5")
    output_fpath = os.path.join(output_dir, output_fname)

    write_hdf(
        pixel_array=cropped_raw,
        fname=output_fpath,
        dataset_name="raw_rescaled",
        compression="gzip"
    )


if __name__ == "__main__":
    RAW_DIR = "/cluster/scratch/kyamauch/full_data_20220301/denoised"
    OVARY_SEGMENTATION_DIR = "/cluster/scratch/kyamauch/full_data_20220301/ovary_segmentation/segmentations"
    OUTPUT_DIR = "/cluster/scratch/kyamauch/full_data_20220301/ovary_segmentation/follicle_segmentation/follicle_datasets"

    MIN_SHAPE = (80, 80, 80)
    MASK_RAW_IM = True

    # get the file names
    raw_file_pattern = os.path.join(RAW_DIR, "*.h5")
    raw_files = glob.glob(raw_file_pattern)

    writer_func = partial(
        create_follicle_dataset,
        segmentation_dir=OVARY_SEGMENTATION_DIR,
        output_dir=OUTPUT_DIR,
        min_shape=MIN_SHAPE,
        mask_raw_im=MIN_SHAPE
    )

    with mp.get_context('spawn').Pool() as pool:
        pool.map(writer_func, raw_files)




from functools import partial
import glob
import os
from typing import Optional, Union, List

import numpy as np
from scipy import ndimage
from scipy.special import softmax
from skimage.measure import regionprops_table
from skimage.morphology import binary_dilation, cube

from ..io.hdf5_io import load_im_from_hdf, write_multi_dataset_hdf, load_raw_im


def post_process_ovary(
    ovary_prediction: np.ndarray,
    raw_im: np.ndarray,
    threshold: float = 0.8,
    dilation_size: Optional[int] = None,
    mask_padding: bool = True,
) -> np.ndarray:
    ovary_mask = np.zeros_like(ovary_prediction, dtype=np.bool)

    # mask out the padding region of the image
    if mask_padding is True:
        ovary_prediction[raw_im == 0] = 0
    ovary_mask[ovary_prediction >= threshold] = True

    # filter for the largest object
    ovary_mask = ndimage.binary_fill_holes(ovary_mask)
    ovary_labels = ndimage.label(ovary_mask)[0]
    region_props = regionprops_table(
        ovary_labels, properties=['label', 'area']
    )
    if len(region_props['label']) > 1:
        max_area_ind = np.argmax(region_props['area'])
        max_label = region_props['label'][max_area_ind]
        ovary_mask[ovary_labels != max_label] = False

    # apply the dilation is requested
    if dilation_size is not None:
        ovary_mask = binary_dilation(ovary_mask, selem=cube(dilation_size))

    return ovary_mask


def segment_ovary_from_prediction(
    prediction_path: Union[str, os.PathLike],
    raw_im_path: Union[str, os.PathLike],
    output_dir: Union[str, os.PathLike],
    threshold: float = 0.8,
    apply_sigmoid: bool = False,
    apply_softmax: bool = False,
    prediction_index: int = 0,
    dilation_size: Optional[int] = None,
    mask_padding: bool = True,
    raw_im_key: str = 'raw_rescaled',
    follicle_labels_key: str = 'follicle_labels_rescaled',
    ovary_labels_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
) -> str:
    """Segment an ovary image from CNN predictions.

    Parameters
    ----------
    prediction_path : Union[str, os.PathLike]
        The path to the predictions file.
    raw_im_path : Union[str, os.PathLike]
        The path to the raw image file. This will be the dataset
        the predictions were made from.
    output_dir : Union[str, os.PathLike]
        The path to the directory the segmentation should be saved to.
    threshold : float
        The probability threshold for binarizing the image.
        The default value is 0.8.
    dilation_size : Optional[int]
        The size of the morphological dilation to apply following binarization
        of the probabilities image. If None is give, no dilation is performed.
        The default value is None.
    raw_im_key : str
        The key in the raw_im file to access to the raw image dataset.
        The default value is raw_rescaled.
    follicle_labels_key : str
        The key in the raw_im file to access to the follicle labels dataset.
        The default value is follicle_labels_rescaled.
    ovary_labels_key : str
        The key in the raw_im file to access to the follicle labels dataset.
        The default value is 'ovary_labels_rescaled'
    prediction_key : str
        The key in the raw_im file to access to the follicle labels dataset.
        The default value is 'predictions'

    Returns
    -------
    output_fpath : str
        The path to the resulting segmentation.
    """
    # load the images
    prediction = load_im_from_hdf(
        fpath=prediction_path, dataset_key=prediction_key
    )
    _, _, raw_rescaled, rescaled_attrs = load_raw_im(
        fpath=raw_im_path, rescaled_name=raw_im_key
    )

    if follicle_labels_key is not None:
        follicle_labels = load_im_from_hdf(
            fpath=raw_im_path, dataset_key=follicle_labels_key
        )
    else:
        follicle_labels = None

    if ovary_labels_key is not None:
        ovary_labels = load_im_from_hdf(
            fpath=raw_im_path, dataset_key=ovary_labels_key
        )
    else:
        ovary_labels = None

    if (apply_sigmoid is True) and (apply_softmax is True):
        raise ValueError('apply_sigmoid and apply_softmax cannot both be True')
    elif apply_sigmoid is True:
        prediction = prediction[prediction_index, ...]
        prediction = 1 / (1 + np.exp(-prediction))
    elif apply_softmax is True:
        prediction = softmax(prediction, axis=0)
        prediction = prediction[prediction_index, ...]

    # segment the ovary
    ovary_segmentation = post_process_ovary(
        ovary_prediction=prediction,
        raw_im=raw_rescaled,
        threshold=threshold,
        dilation_size=dilation_size,
        mask_padding=mask_padding,
    )

    # save the segmentation
    output_fname = os.path.basename(prediction_path).replace(
        '_predictions.', '_segmentation.'
    )
    output_fpath = os.path.join(output_dir, output_fname)
    rescaled_dataset = {'data': raw_rescaled, 'attrs': rescaled_attrs}

    if (follicle_labels is not None) and (ovary_labels is not None):
        write_multi_dataset_hdf(
            file_path=output_fpath,
            compression='gzip',
            raw_rescaled=rescaled_dataset,
            follicle_labels_rescaled=follicle_labels,
            ovary_labels_rescaled=ovary_labels,
            ovary_segmentation_rescaled=ovary_segmentation,
        )
    else:
        write_multi_dataset_hdf(
            file_path=output_fpath,
            compression='gzip',
            raw_rescaled=rescaled_dataset,
            ovary_segmentation_rescaled=ovary_segmentation,
        )

    return output_fpath


def segment_ovary_dir(
    dir_path: Union[str, os.PathLike],
    raw_im_dir: Union[str, os.PathLike],
    predictions_subdir: str,
    output_dir_name: Union[str, os.PathLike],
    threshold: float = 0.8,
    apply_sigmoid: bool = False,
    apply_softmax: bool = False,
    prediction_index: int = 0,
    dilation_size: Optional[int] = None,
    mask_padding: bool = True,
    raw_im_key: str = 'raw_rescaled',
    follicle_labels_key: str = 'follicle_labels_rescaled',
    ovary_labels_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
) -> List[str]:
    predictions_dir = os.path.join(dir_path, predictions_subdir)
    predictions_pattern = os.path.join(predictions_dir, '*_predictions.h5')
    prediction_files = glob.glob(predictions_pattern)

    output_dir_fpath = os.path.join(dir_path, output_dir_name)
    if not os.path.isdir(output_dir_fpath):
        os.mkdir(output_dir_fpath)

    segmentation_fpaths = []
    for prediction in prediction_files:
        raw_im_name = os.path.basename(prediction).replace(
            '_predictions.', '.'
        )
        raw_im_path = os.path.join(raw_im_dir, raw_im_name)
        seg_path = segment_ovary_from_prediction(
            prediction_path=prediction,
            raw_im_path=raw_im_path,
            output_dir=output_dir_fpath,
            threshold=threshold,
            dilation_size=dilation_size,
            mask_padding=mask_padding,
            apply_sigmoid=apply_sigmoid,
            apply_softmax=apply_softmax,
            prediction_index=prediction_index,
            raw_im_key=raw_im_key,
            follicle_labels_key=follicle_labels_key,
            ovary_labels_key=ovary_labels_key,
            prediction_key=prediction_key,
        )

        segmentation_fpaths.append(seg_path)

    return segmentation_fpaths


def post_process_ovary_cross_val(
    crossval_dir: Union[str, os.PathLike],
    ovary_datasets_dir: Union[str, os.PathLike],
    predictions_subdir: str,
    output_dir_name: Union[str, os.PathLike],
    threshold: float = 0.8,
    apply_sigmoid: bool = False,
    apply_softmax: bool = False,
    prediction_index: int = 0,
    dilation_size: Optional[int] = None,
    mask_padding: bool = True,
    raw_im_key: str = 'raw_rescaled',
    follicle_labels_key: str = 'follicle_labels_rescaled',
    ovary_labels_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
):
    # get the fold directories
    folder_pattern = os.path.join(crossval_dir, 'fold_*', '')
    fold_candidates = glob.glob(folder_pattern)

    # make sure they are directories
    fold_dirs = [f for f in fold_candidates if os.path.isdir(f)]
    n_fold = len(fold_dirs)
    print(f'processing {n_fold} folds')

    # make the base function
    process_func = partial(
        segment_ovary_dir,
        raw_im_dir=ovary_datasets_dir,
        predictions_subdir=predictions_subdir,
        output_dir_name=output_dir_name,
        threshold=threshold,
        apply_sigmoid=apply_sigmoid,
        apply_softmax=apply_softmax,
        prediction_index=prediction_index,
        dilation_size=dilation_size,
        mask_padding=mask_padding,
        raw_im_key=raw_im_key,
        follicle_labels_key=follicle_labels_key,
        ovary_labels_key=ovary_labels_key,
        prediction_key=prediction_key,
    )

    for folder in fold_dirs:
        process_func(folder)



import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt


def make_cycle_plot(
    follicle_data: pd.DataFrame,
    x_axis: str,
    y_axis: str,
    style_key: bool = None,
    plot_type: str = 'strip',
    hue: str = 'side',
    rotate_xlabel: bool = False,
    figsize: tuple = (20, 9),
    fontsize: int = 12,
):
    """plotting function which includes strip, line, scatter, box plots

    Parameters
    ----------
    follicle_data: pandas Dataframe
        The data needed to be plotted
    yaxis: str
        the name of column which include the values that are going to be used.
    xaxis: str
        variable of x axis such as relative day or day from peak
    style_key: str
        the style option.
    plot_type: str
        the plot type such as line, scatter, strip or box
    hue: str
        the way to separate data, such as side
    figsize: tuple
        the size of figure

    Returns
    -------
    f: Figure
        the plot
    axs: axes.Axes or array of Axes
        ax can be either a single Axes object or an array of Axes objects if more than one subplot was created.

    """
    f, axs = plt.subplots(1, 2, sharey=True, figsize=figsize)

    base_kwargs = {'x': x_axis, 'y': y_axis, 'hue': hue}
    if style_key is not None:
        base_kwargs.update({'style': style_key})

    z1_data = follicle_data.loc[follicle_data['cycle'] == 'z1']
    z1_kwargs = base_kwargs
    z1_kwargs.update({'data': z1_data, 'ax': axs[0]})
    if plot_type == 'strip':
        sns.stripplot(**z1_kwargs)
    elif plot_type == 'line':
        z1_kwargs.update({'legend': 'brief'})
        sns.lineplot(**z1_kwargs)
    elif plot_type == 'scatter':
        z1_kwargs.update({'alpha': 1})
        sns.scatterplot(**z1_kwargs)
    elif plot_type == 'box_strip':
        sns.stripplot(x=x_axis, y=y_axis, data=z1_data, ax=axs[0], color='.25')
        z1_kwargs.update({'palette': "Set3"})
        sns.boxplot(**z1_kwargs)
    elif plot_type == 'box':
        z1_kwargs.update({'palette': "Set3"})
        sns.boxplot(**z1_kwargs)

    z2_data = follicle_data.loc[follicle_data['cycle'] == 'z2']
    z2_kwargs = base_kwargs
    z2_kwargs.update({'data': z2_data, 'ax': axs[1]})
    if plot_type == 'strip':
        sns.stripplot(**z2_kwargs)
    elif plot_type == 'line':
        z1_kwargs.update({'legend': 'brief'})
        sns.lineplot(**z2_kwargs)
    elif plot_type == 'scatter':
        z2_kwargs.update({'alpha': 1})
        sns.scatterplot(**z2_kwargs)
    elif plot_type == 'box_strip':
        sns.stripplot(x=x_axis, y=y_axis, data=z2_data, ax=axs[1], color='.25')
        z2_kwargs.update({'palette': "Set3"})
        sns.boxplot(**z2_kwargs)
    elif plot_type == 'box':
        z2_kwargs.update({'palette': "Set3"})
        sns.boxplot(**z2_kwargs)

    # remove the right legend
    right_legend = axs[1].legend()
    right_legend.remove()

    # add cycle titles
    axs[0].set_xlabel('cycle 1', fontsize=fontsize)
    axs[1].set_xlabel('cycle 2', fontsize=fontsize)

    if rotate_xlabel:
        # rotate label
        axs[0].set_xticklabels(
            axs[0].get_xticklabels(),
            rotation=40,
            ha="right",
            fontsize=fontsize,
        )
        axs[1].set_xticklabels(
            axs[1].get_xticklabels(),
            rotation=40,
            ha="right",
            fontsize=fontsize,
        )

    # hide the vertical axis on the right plot
    axs[1].get_yaxis().set_visible(False)

    # remove the box
    axs[0].spines['right'].set_visible(False)
    axs[0].spines['top'].set_visible(False)

    axs[1].spines['left'].set_visible(False)
    axs[1].spines['right'].set_visible(False)
    axs[1].spines['top'].set_visible(False)

    f.tight_layout()

    return f, axs



from ovary_analysis.post_process.follicles import segment_follicles_fold

if __name__ == "__main__":
    FOLD_DIR_PATH = "/cluster/work/cobi/kevin/ivf/cross_validation_20220126/follicle_crossval_mask/fold_0"
    PREDICTIONS_DIR_NAME = "predictions"
    OUTPUT_DIR_NAME = 'all_segmentations'
    PROB_THRESHOLD = 0.5
    VOLUME_THRESHOLD = 30
    PREDICTION_INDEX = 1
    APPLY_SOFTMAX = True
    RAW_IM_KEY = 'raw_rescaled'
    FOLLICLES_LABELS_KEY = 'follicle_labels_rescaled'
    OVARY_SEGMENTATION_KEY = 'ovary_labels_rescaled'
    PREDICTIONS_KEY = 'predictions'

    for folder in ["test", "train", "val"]:
        segment_follicles_fold(
            fold_dir_path=FOLD_DIR_PATH,
            raw_im_dir_name=folder,
            output_dir_name=OUTPUT_DIR_NAME,
            predictions_dir_name=PREDICTIONS_DIR_NAME,
            prob_threshold=PROB_THRESHOLD,
            volume_threshold=VOLUME_THRESHOLD,
            prediction_index=PREDICTION_INDEX,
            apply_softmax=APPLY_SOFTMAX,
            raw_im_key=RAW_IM_KEY,
            follicle_labels_key=FOLLICLES_LABELS_KEY,
            ovary_segmentation_key=OVARY_SEGMENTATION_KEY,
            prediction_key=PREDICTIONS_KEY,
        )



import pytest

from ovary_analysis.utils.file_names import is_cycle_folder

valid_folder_paths = [
    '/test/z1Links',
    '/test/z2Links',
    '/test/z1Rechts',
    '/test/z2Rechts',
    '/test/Z1 Links',
    '/test/Z2 Links',
    '/test/Z1 Rechts',
    '/test/Z2 Rechts',
    './z1Links',
    './Z1 Links',
    './Z2 Links',
    './Z1 Rechts',
    './Z2 Rechts',
]


@pytest.mark.parametrize('folder_path', valid_folder_paths)
def test_valid_cycle_folder(folder_path):
    is_valid = is_cycle_folder(folder_path)
    assert is_valid is True


invalid_folder_paths = [
    '/test/z1',
    '/test/Links',
    '/test/z2',
    '/test/rechts',
    '/test/z1/links',
    '/test/z1z2links',
    '/test/z1z2rechtslinks',
]


@pytest.mark.parametrize('folder_path', invalid_folder_paths)
def test_invalid_cycle_folder(folder_path):
    is_valid = is_cycle_folder(folder_path)
    assert is_valid is False



"""
This example creates a dataframe for screening
effect of probability threshold on IOU
"""
from ovary_analysis.post_process.segmentation_metrics import measure_iou_from_crossval


thresholds = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0,9]
crossval_dir = '/cluster/work/cobi/kevin/ivf/cross_validation_20220126/ovary_cross_validation'
measurement_table = measure_iou_from_crossval(
    crossval_dir=crossval_dir,
    fold_dir_pattern='fold_*',
    ref_dir='test',
    prediction_dir='predictions',
    thresholds=thresholds,
    label_key='ovary_labels_rescaled',
    prediction_key='predictions'
)

measurement_table.to_csv('ovary_ious.csv')



import datetime
import os
from typing import Optional

import h5py
import numpy as np
import pandas as pd
from skimage.morphology import binary_dilation, binary_erosion, cube
from skimage.segmentation import find_boundaries

from ..io.hdf5_io import write_training_data
from ..post_process.ovary import post_process_ovary
from ..utils.time import date_to_datestring


def crop_image(image, mask_image, min_shape=[80, 80, 80]):
    # get the crop extents
    mask_coords = np.argwhere(mask_image)
    min_indices = np.min(mask_coords, axis=0)
    max_indices = np.max(mask_coords, axis=0) + 1

    crop_shape = max_indices - min_indices
    if not np.all(crop_shape >= min_shape):
        # if the crop is smaller than the specified minimum size, expand it
        min_acceptable_indices = min_indices + min_shape
        max_indices[crop_shape < min_shape] = min_acceptable_indices[
            crop_shape < min_shape
        ]

    # get the crop
    cropped_image = image[
        min_indices[0] : max_indices[0],
        min_indices[1] : max_indices[1],
        min_indices[2] : max_indices[2],
    ]

    return cropped_image


def crop_raw_image(
    raw_image, ovary_image, min_shape=[80, 80, 80], mask_periphery: bool = True
):
    ovary_mask = ovary_image.astype(np.bool)

    # get the average pixel value and set all non ovary pixels
    if mask_periphery is True:
        mean_px_value = raw_image.mean()
        raw_image[np.logical_not(ovary_mask)] = mean_px_value

    cropped_raw_image = crop_image(raw_image, ovary_image, min_shape=min_shape)

    return cropped_raw_image


def crop_label_image(label_image, ovary_image, min_shape=[80, 80, 80]):
    ovary_mask = ovary_image.astype(np.bool)
    cropped_label = crop_image(label_image, ovary_mask, min_shape=min_shape)

    return cropped_label


def create_follicle_weights(
    ovary_im: np.ndarray,
    follicles_im: np.ndarray,
    padding_weight: float = 0,
    ovary_weight: float = 0.8,
    follicle_weight: float = 0.8,
    boundary_weight: float = 1,
    follicle_dilation: int = 4,
    follicle_erosion: int = 2,
):
    ovary_mask = ovary_im.astype(np.bool)
    follicles_mask = follicles_im.astype(np.bool)

    weights = np.zeros_like(ovary_mask, dtype=np.float)

    # apply the padding weights
    weights[np.logical_not(ovary_mask)] = padding_weight

    # apply the ovary weights
    weights[ovary_mask] = ovary_weight

    # apply the follicle weights
    weights[follicles_mask] = follicle_weight

    # apply the follicle boundary weights
    dilated_follicles = binary_dilation(
        follicles_mask, selem=cube(follicle_dilation)
    )
    eroded_follicles = binary_erosion(
        follicles_mask, selem=cube(follicle_erosion)
    )
    boundary_mask = np.logical_xor(dilated_follicles, eroded_follicles)
    weights[boundary_mask] = boundary_weight

    return weights


def create_multiclass_labels(follicle_labels: np.ndarray) -> np.ndarray:
    follicles_mask = follicle_labels.astype(bool)
    multiclass_labels = np.zeros_like(follicle_labels, dtype=np.int)
    multiclass_labels[follicles_mask] = 1

    # create periphery region
    follicle_boundaries = find_boundaries(follicles_mask, mode='thick')

    # set labels for periphery
    multiclass_labels[follicle_boundaries] = 2

    multiclass_one_hot = np.stack(
        [follicles_mask, follicle_boundaries]
    ).astype(np.uint8)

    return multiclass_labels, multiclass_one_hot


def create_follicle_dataset(
    ovary_prediction_fname: str,
    ovary_dataset_fname: str,
    output_directory: str,
    patient_id: str,
    cycle_id: str,
    side: str,
    day: datetime.date,
    mask_raw_im: bool = True,
    min_shape=[80, 80, 80],
    raw_name: str = 'raw',
    follicle_name: str = 'follicle_labels',
    ovary_name: str = 'ovary_labels',
    weights_name: str = 'weights',
    threshold: float = 0.5,
    dilation_size: int = 10,
    follicle_dilation: int = 4,
    follicle_erosion: int = 2,
    ovary_weight: float = 0.8,
    follicle_weight: float = 0.8,
    boundary_weight: float = 1,
    multiclass_name: Optional[str] = None,
):

    # load the ovary prediction
    with h5py.File(ovary_prediction_fname) as f_prediction:
        ovary_prediction = f_prediction['predictions'][1, :, :, :]

    with h5py.File(ovary_dataset_fname) as f_ovary:
        raw_im = f_ovary[raw_name][:]
        follicle_im = f_ovary[follicle_name][:]

    # make sure the follicle labels are 0 or 1
    follicle_im[follicle_im > 0] = 1
    follicle_im[follicle_im == 0] = 0

    # create the ovary mask
    ovary_mask = post_process_ovary(
        raw_im=raw_im,
        ovary_prediction=ovary_prediction,
        threshold=threshold,
        dilation_size=dilation_size,
    )

    # crop the data
    cropped_raw = crop_raw_image(
        raw_im, ovary_mask, min_shape=min_shape, mask_periphery=mask_raw_im
    )
    cropped_follicle = crop_label_image(
        follicle_im, ovary_mask, min_shape=min_shape
    )
    cropped_ovary = crop_label_image(
        ovary_mask.copy(), ovary_mask, min_shape=min_shape
    )

    # create the weights image
    weights = create_follicle_weights(
        ovary_im=cropped_ovary,
        follicles_im=cropped_follicle,
        padding_weight=0,
        ovary_weight=ovary_weight,
        follicle_weight=follicle_weight,
        boundary_weight=boundary_weight,
        follicle_dilation=follicle_dilation,
        follicle_erosion=follicle_erosion,
    )

    # create the filename
    date_string = date_to_datestring(day)
    fname = f'{patient_id}_{cycle_id}_{side}_{date_string}_follicles.h5'
    file_path = os.path.join(output_directory, fname)

    if multiclass_name is not None:
        multiclass_im, multiclass_one_hot = create_multiclass_labels(
            cropped_follicle
        )

    else:
        multiclass_im = None
        multiclass_one_hot = None
    # write the data
    write_training_data(
        raw_im=cropped_raw,
        raw_im_path=ovary_dataset_fname,
        follicle_im=cropped_follicle,
        follicle_im_path=ovary_dataset_fname,
        ovary_im=cropped_ovary,
        ovary_im_path=ovary_prediction_fname,
        file_path=file_path,
        raw_name=raw_name,
        follicle_name=follicle_name,
        ovary_name=ovary_name,
        weights_im=weights,
        weights_name=weights_name,
        multiclass_im=multiclass_im,
        multiclass_name=multiclass_name,
        multiclass_one_hot_im=multiclass_one_hot,
        compression='gzip',
    )

    return file_path, True


def batch_create_follicle_dataset(
    dataset_table: pd.DataFrame,
    ovary_datasets_dir: str,
    ovary_predictions_dir: str,
    output_base_dir: str,
    raw_name: str = 'raw',
    follicle_name: str = 'follicle_labels',
    ovary_name: str = 'ovary_labels',
    weights_name: str = 'weights',
    mask_raw_im: bool = True,
    min_shape=[80, 80, 80],
    threshold: float = 0.5,
    dilation_size: int = 10,
    follicle_dilation: int = 4,
    follicle_erosion: int = 2,
    ovary_weight: float = 0.8,
    follicle_weight: float = 0.8,
    boundary_weight: float = 1,
    multiclass_name: Optional[str] = None,
):

    # create the output directories
    group_names = dataset_table['group'].unique()
    for group in group_names:
        os.mkdir(os.path.join(output_base_dir, group))

    file_paths = []
    success = []

    for _, row in dataset_table.iterrows():
        patient_id = row['patient']
        cycle_id = row['cycle']
        side = row['side']
        day = row['day']
        group = row['group']

        if isinstance(day, str):
            date_time_obj = datetime.datetime.strptime(day, '%Y-%m-%d')
            date_string = date_time_obj.strftime('%y%m%d')
        else:
            date_string = day.strftime('%y%m%d')
        filename_base = f'{patient_id}_{cycle_id}_{side}_{date_string}_ovary'

        ovary_prediction_fname = os.path.join(
            ovary_predictions_dir, filename_base + '_predictions.h5'
        )
        ovary_dataset_fname = os.path.join(
            ovary_datasets_dir, group, filename_base + '.h5'
        )
        output_directory = os.path.join(output_base_dir, group)

        file_path, conversion_succeeded = create_follicle_dataset(
            ovary_prediction_fname=ovary_prediction_fname,
            ovary_dataset_fname=ovary_dataset_fname,
            output_directory=output_directory,
            patient_id=patient_id,
            cycle_id=cycle_id,
            side=side,
            day=day,
            raw_name=raw_name,
            follicle_name=follicle_name,
            ovary_name=ovary_name,
            weights_name=weights_name,
            mask_raw_im=mask_raw_im,
            min_shape=min_shape,
            threshold=threshold,
            dilation_size=dilation_size,
            ovary_weight=ovary_weight,
            follicle_weight=follicle_weight,
            boundary_weight=boundary_weight,
            follicle_dilation=follicle_dilation,
            follicle_erosion=follicle_erosion,
            multiclass_name=multiclass_name,
        )

        file_paths.append(file_path)
        success.append(conversion_succeeded)

    dataset_table['follicle_path'] = file_paths
    dataset_table['follicle_conversion_successful'] = success

    return dataset_table



from functools import partial
import glob
import multiprocessing as mp
import os
from typing import Tuple, Union

import h5py

from ovary_analysis.io.hdf5_io import load_im_from_hdf, write_multi_dataset_hdf
from ovary_analysis.post_process.follicles import post_process_follicles
from ovary_analysis.utils.image import crop_label_image


def segment_follicles(
    prediction_fpath: str,
    raw_im_dir_path: str,
    ovary_segmentation_dir_path: Union[str, os.PathLike],
    output_dir_path: Union[str, os.PathLike],
    prob_threshold: float,
    volume_threshold: float,
    prediction_index: int = 1,
    apply_softmax: bool = True,
    raw_im_key: str = 'raw_rescaled',
    ovary_segmentation_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
    min_shape: Tuple[int, int, int] = (80, 80, 80)
):

    if not os.path.isdir(output_dir_path):
        os.mkdir(output_dir_path)

    # load the raw image
    prediction_fname = os.path.basename(prediction_fpath)
    raw_fname = prediction_fname.replace("_predictions.h5", ".h5")
    raw_im_fpath = os.path.join(raw_im_dir_path, raw_fname)
    with h5py.File(raw_im_fpath, 'r') as f_raw:
        raw_rescaled = f_raw[raw_im_key][:]
        rescaled_attrs = dict(f_raw[raw_im_key].attrs.items())

    # get the ovary segmentation image path
    ovary_segmentation_fname = prediction_fname.replace(
        "_follicles_predictions.h5",
        "_raw_segmentation.h5"
    )
    ovary_segmentation_fpath = os.path.join(ovary_segmentation_dir_path, ovary_segmentation_fname)

    # load the ovary segmentation
    prediction = load_im_from_hdf(
        fpath=prediction_fpath, dataset_key=prediction_key
    )
    ovary_segmentation = load_im_from_hdf(
        fpath=ovary_segmentation_fpath, dataset_key=ovary_segmentation_key
    )
    cropped_ovary_segmentation = crop_label_image(
        label_image=ovary_segmentation,
        ovary_image=ovary_segmentation,
        min_shape=min_shape
    )

    # segment the follicles
    follicles_segmentation = post_process_follicles(
        follicle_prediction=prediction,
        ovary_seg=cropped_ovary_segmentation.astype(bool),
        prob_threshold=prob_threshold,
        volume_threshold=volume_threshold,
        prediction_index=prediction_index,
        apply_softmax=apply_softmax,
    )

    # save the output
    output_fname = os.path.basename(prediction_fpath).replace(
        '_predictions.', '_segmentation.'
    )
    output_fpath = os.path.join(output_dir_path, output_fname)
    rescaled_dataset = {'data': raw_rescaled, 'attrs': rescaled_attrs}
    write_multi_dataset_hdf(
        file_path=output_fpath,
        compression="gzip",
        raw_rescaled=rescaled_dataset,
        follicle_segmentation_rescaled=follicles_segmentation,
        ovary_segmentation_rescaled=cropped_ovary_segmentation
    )

    return output_fpath


if __name__ == "__main__":
    PREDICTIONS_DIR_PATH = "/cluster/scratch/kyamauch/full_data_20220301/follicle_segmentation/predictions"
    RAW_IM_DIR_PATH = "/cluster/scratch/kyamauch/full_data_20220301/follicle_segmentation/follicle_datasets"
    OVARY_SEGMENTATION_DIR_PATH = "/cluster/scratch/kyamauch/full_data_20220301/ovary_segmentation/segmentations"
    OUTPUT_DIR_PATH = "/cluster/scratch/kyamauch/full_data_20220301/follicle_segmentation/segmentations"

    PREDICTIONS_DIR_NAME = 'predictions'
    OUTPUT_DIR_NAME = 'segmentations'
    PROB_THRESHOLD = 0.5
    VOLUME_THRESHOLD = 30
    PREDICTION_INDEX = 1
    APPLY_SOFTMAX = True
    RAW_IM_KEY = 'raw_rescaled'
    FOLLICLES_LABELS_KEY = 'follicle_labels_rescaled'
    OVARY_SEGMENTATION_KEY = 'ovary_labels_rescaled'
    PREDICTIONS_KEY = 'predictions'

    segmentation_function = partial(
        segment_follicles,
        raw_im_dir_path=RAW_IM_DIR_PATH,
        ovary_segmentation_dir_path=OVARY_SEGMENTATION_DIR_PATH,
        output_dir_path=OUTPUT_DIR_PATH,
        prob_threshold=PROB_THRESHOLD,
        volume_threshold=VOLUME_THRESHOLD,
        prediction_index=PREDICTION_INDEX,
        apply_softmax=PREDICTION_INDEX,
        raw_im_key=RAW_IM_KEY,
        follicle_labels_key=FOLLICLES_LABELS_KEY,
        ovary_segmentation_key=OVARY_SEGMENTATION_KEY,
        prediction_key=PREDICTIONS_KEY,
    )

    prediction_paths = glob.glob(os.path.join(PREDICTIONS_DIR_PATH, "*.h5"))

    with mp.get_context('spawn').Pool() as pool:
        pool.map(segmentation_function, prediction_paths)



from ovary_analysis.post_process.ovary import segment_ovary_dir

BASE_DIR = "/cluster/work/cobi/kevin/ivf/full_data_20220301/ovary_segmentation"
RAW_IM_DIR = "/cluster/work/cobi/kevin/ivf/full_data_20220301/raw_data/denoised"
PROBABILITY_THRESHOLD = 0.5
APPLY_SIGMOID = False
APPLY_SOFTMAX = True
DILATION_SIZE = None
MASK_PADDING = False


segment_ovary_dir(
    dir_path=BASE_DIR,
    raw_im_dir=RAW_IM_DIR,
    predictions_subdir='predictions',
    output_dir_name='segmentations',
    threshold=PROBABILITY_THRESHOLD,
    apply_sigmoid=APPLY_SIGMOID,
    apply_softmax=APPLY_SOFTMAX,
    prediction_index=1,
    dilation_size=DILATION_SIZE,
    mask_padding=MASK_PADDING,
    raw_im_key='raw_rescaled',
    follicle_labels_key=None,
    ovary_labels_key=None,
    prediction_key='predictions',
)



import pandas as pd
from pandas._testing import assert_frame_equal
import numpy as np
from ovary_analysis.utils import measurement_table


def test_calc_relative_day():
    # Setup
    valid_DataFrame = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A02', 'A02'],
            'side': ['links', 'links', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z2', 'z2'],
            'date': ['2017/4/3', '2017/4/5', '2017/5/17', '2017/5/19'],
            'volume': [10, 20, 30, 50],
        }
    )
    valid_DataFrame['date'] = pd.to_datetime(valid_DataFrame['date'])
    valid_DataFrame['day_rel'] = np.nan

    valid_df_rel = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A02', 'A02'],
            'side': ['links', 'links', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z2', 'z2'],
            'date': ['2017/4/3', '2017/4/5', '2017/5/17', '2017/5/19'],
            'volume': [10, 20, 30, 50],
            'day_rel': [0, 2, 0, 2],
        }
    )
    valid_df_rel['date'] = pd.to_datetime(valid_df_rel['date'])

    # Exercise
    df_rel = measurement_table.calc_relative_day(valid_DataFrame)

    # Verify
    assert_frame_equal(df_rel, valid_df_rel, check_dtype=False)

    # Cleanup - none necessary


def test_calc_growth():
    # Setup
    df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A02', 'A02'],
            'side': ['links', 'links', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z2', 'z2'],
            'date': ['2017/4/3', '2017/4/5', '2017/5/17', '2017/5/19'],
            'volume': [10, 20, 30, 50],
            'day_rel': [0, 2, 0, 2],
        }
    )

    valid_df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A02', 'A02'],
            'side': ['links', 'links', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z2', 'z2'],
            'date': ['2017/4/3', '2017/4/5', '2017/5/17', '2017/5/19'],
            'volume': [10, 20, 30, 50],
            'day_rel': [0, 2, 0, 2],
            'growth': [0, 10, 0, 20],
        }
    )
    # Exercise
    df1 = measurement_table.calc_growth(df)

    # Verify
    assert_frame_equal(df1, valid_df, check_dtype=False)

    # Cleanup - none necessary


def test_extract_dominant():
    # Setup
    df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A02', 'A02'],
            'side': ['links', 'links', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z2', 'z2'],
            'date': ['2017/4/3', '2017/4/3', '2017/5/17', '2017/5/17'],
            'volume': [10, 20, 30, 50],
            'day_rel': [0, 0, 0, 0],
        }
    )
    valid_df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A02'],
            'side': ['links', 'rechts'],
            'cycle': ['z1', 'z2'],
            'date': ['2017/4/3', '2017/5/17'],
            'volume': [20, 50],
            'day_rel': [0, 0],
            'original_index': [1, 3],
        }
    )

    # Exercise
    df1 = measurement_table.extract_dominant(df, level='cycle')

    # Verify
    assert_frame_equal(df1, valid_df, check_dtype=False, check_like=True)

    # Cleanup - none necessary


def test_voxel_to_mm():
    # Setup
    df = pd.DataFrame({'volume': [2, 5]})
    valid_df = pd.DataFrame({'volume': [0.0077398, 0.019350]})

    # Exercise
    df1 = measurement_table.voxel_to_mm(df)

    # Verify
    assert_frame_equal(
        df1, valid_df, check_dtype=False, check_like=True, check_less_precise=5
    )

    # Cleanup - none necessary


def test_remove_dominant_follicle():
    # Setup
    df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A02', 'A02'],
            'side': ['links', 'links', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z2', 'z2'],
            'date': ['2017/4/3', '2017/4/3', '2017/5/17', '2017/5/17'],
            'volume': [10, 20, 30, 50],
            'day_rel': [0, 0, 0, 0],
        }
    )
    valid_df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A02'],
            'side': ['links', 'rechts'],
            'cycle': ['z1', 'z2'],
            'date': ['2017/4/3', '2017/5/17'],
            'volume': [10, 30],
            'day_rel': [0, 0],
        }
    )
    # Exercise
    df1 = measurement_table.remove_dominant_follicle(df)

    # Verify
    assert_frame_equal(df1, valid_df, check_dtype=False, check_like=True)

    # Cleanup - none necessary


def test_unique_patient():
    # Setup
    df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A02', 'A02'],
            'side': ['links', 'links', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z2', 'z2'],
            'date': ['2017/4/3', '2017/4/3', '2017/5/17', '2017/5/17'],
            'volume': [10, 20, 30, 50],
            'day_rel': [0, 0, 1, 1],
        }
    )
    valid_df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A02'],
            'side': ['links', 'rechts'],
            'cycle': ['z1', 'z2'],
            'date': ['2017/4/3', '2017/5/17'],
            'volume': [20, 50],
            'day_rel': [0, 1],
        }
    )

    # Exercise
    df1 = measurement_table.unique_patient(df)

    # Verify
    assert_frame_equal(df1, valid_df, check_dtype=False, check_like=True)

    # Cleanup - none necessary


def test_calc_domi_relative_day():
    # Setup
    df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A01', 'A02', 'A02', 'A02'],
            'side': ['links', 'links', 'links', 'rechts', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z1', 'z2', 'z2', 'z2'],
            'date': [
                '2017/4/3',
                '2017/4/5',
                '2017/4/5',
                '2017/5/17',
                '2017/5/19',
                '2017/5/21',
            ],
            'volume': [10, 20, 25, 30, 50, 45],
            'day_rel': [0, 2, 2, 0, 2, 4],
        }
    )
    valid_df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A01', 'A02', 'A02', 'A02'],
            'side': ['links', 'links', 'links', 'rechts', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z1', 'z2', 'z2', 'z2'],
            'date': [
                '2017/4/3',
                '2017/4/5',
                '2017/4/5',
                '2017/5/17',
                '2017/5/19',
                '2017/5/21',
            ],
            'volume': [10, 20, 25, 30, 50, 45],
            'day_rel': [0, 2, 2, 0, 2, 4],
            'day_from_peak': [-2, 0, 0, -2, 0, 2],
        }
    )

    # Exercise
    df1 = measurement_table.calc_domi_relative_day(df)

    # Verify
    assert_frame_equal(df1, valid_df, check_dtype=False, check_like=True)


def test_group_sparse():
    # Setup
    df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A01', 'B02', 'B02', 'B02'],
            'side': ['links', 'links', 'links', 'rechts', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z1', 'z2', 'z2', 'z2'],
            'date': [
                '2017/4/3',
                '2017/4/5',
                '2017/4/5',
                '2017/5/17',
                '2017/5/19',
                '2017/5/21',
            ],
            'volume': [10, 20, 25, 30, 50, 45],
            'day_rel': [0, 2, 2, 0, 2, 4],
        }
    )
    valid_df = pd.DataFrame(
        {
            'patient_id': ['A01', 'A01', 'A01', 'B02', 'B02', 'B02'],
            'side': ['links', 'links', 'links', 'rechts', 'rechts', 'rechts'],
            'cycle': ['z1', 'z1', 'z1', 'z2', 'z2', 'z2'],
            'date': [
                '2017/4/3',
                '2017/4/5',
                '2017/4/5',
                '2017/5/17',
                '2017/5/19',
                '2017/5/21',
            ],
            'volume': [10, 20, 25, 30, 50, 45],
            'day_rel': [0, 2, 2, 0, 2, 4],
            'group': ['A', 'A', 'A', 'B', 'B', 'B'],
        }
    )

    # Exercise
    df1 = measurement_table.group_parse(df)

    # Verify
    assert_frame_equal(df1, valid_df, check_dtype=False, check_like=True)



import os

import numpy as np
from skimage.io import imread


def load_label_im(label_path: os.PathLike) -> np.ndarray:
    """Load a label image (tif) into a numpy array

    Parameters
    ----------
    label_path : os.PathLike
        Path to the label image. Should be a tif.
        The labels file is assumed to be in ZCXY order.

    Returns
    -------
    im_array : np.ndarray
        The label image as a numpy array. Ordered CZXY.
    """
    im = imread(label_path)
    im_array = np.swapaxes(im, 1, 0)

    return im_array



from typing import Tuple, Union, List

import numpy as np
from skimage.transform import rescale


def rescale_raw_im(
    raw_im: np.ndarray,
    im_px_size: Tuple[float, float, float],
    target_px_size: Tuple[float, float, float],
) -> np.ndarray:

    scaling_factor = np.divide(im_px_size, target_px_size)
    if np.any(scaling_factor < 1):
        anti_alias = True
    else:
        anti_alias = False
    rescaled_im = rescale(
        raw_im, scaling_factor, anti_aliasing=anti_alias, order=1
    )

    return rescaled_im


def rescale_label_im(
    raw_im: np.ndarray,
    im_px_size: Tuple[float, float, float],
    target_px_size: Tuple[float, float, float],
) -> np.ndarray:

    scaling_factor = np.divide(im_px_size, target_px_size)
    rescaled_im = rescale(raw_im, scaling_factor, order=0)

    return rescaled_im


def pad_label_im(raw_im: np.ndarray, label_im: np.ndarray) -> np.ndarray:
    """Pad a label image to match the shape along the 0th axis
    of the raw image.

    This is a hack to account for the small
    subset of images that have a z dim off by one.

    Parameters
    ----------
    raw_im : np.ndarray
        The raw image. This is the template image.
    label_im : np.ndarray
        This is the image that has its 0th axis padded.

    Returns
    -------
    label_im : np.ndarray
        The padded label image. If no padding is required (i.e., the shape
        of the raw and label images matches), then the original label_im
        is returned.
    """
    if raw_im.shape[0] != label_im.shape[0]:
        n_slices = raw_im.shape[0] - label_im.shape[0]
        assert n_slices > 0
        padding = np.zeros((n_slices, label_im.shape[1], label_im.shape[2]))
        label_im = np.concatenate((label_im, padding), axis=0)
    return label_im


def crop_image(image, mask_image, min_shape=[80, 80, 80]):
    # get the crop extents
    mask_coords = np.argwhere(mask_image)
    min_indices = np.min(mask_coords, axis=0)
    max_indices = np.max(mask_coords, axis=0) + 1

    crop_shape = max_indices - min_indices
    if not np.all(crop_shape >= min_shape):
        # if the crop is smaller than the specified minimum size, expand it
        min_acceptable_indices = min_indices + min_shape
        max_indices[crop_shape < min_shape] = min_acceptable_indices[
            crop_shape < min_shape
        ]

    # get the crop
    cropped_image = image[
        min_indices[0] : max_indices[0],
        min_indices[1] : max_indices[1],
        min_indices[2] : max_indices[2],
    ]

    return cropped_image, image.shape, min_indices


def crop_raw_image(
    raw_image, mask_image, min_shape=[80, 80, 80], mask_periphery: bool = True
):
    ovary_mask = mask_image.astype(bool)

    # get the average pixel value and set all non ovary pixels
    if mask_periphery is True:
        mean_px_value = raw_image.mean()
        raw_image[np.logical_not(ovary_mask)] = mean_px_value

    cropped_raw_image, reference_shape, offset = crop_image(
        raw_image, mask_image, min_shape=min_shape
    )

    return cropped_raw_image, reference_shape, offset


def pad_image(image: np.ndarray, reference_shape: Tuple, offsets: Tuple):
    """
    Pad an image to the reference shape and positioned to the given offset.

    Parameters
    ----------
    image : np.ndarray
        Image to be padded
    reference_shape : Tuple
        Shape of the padded array
    offsets : Tuple
        Offset to set the original array into the padded padded output image
    Returns
    -------
    np.ndarray
        Padded image
    """
    # Create an array of zeros with the reference shape
    result = np.zeros(reference_shape)
    # Create a list of slices from offset to offset + shape in each dimension
    insertHere = [
        slice(offsets[dim], offsets[dim] + image.shape[dim])
        for dim in range(image.ndim)
    ]
    # Insert the array in the result at the specified offsets
    result[tuple(insertHere)] = image
    return result


def crop_label_image(label_image, ovary_image, min_shape=[80, 80, 80]):
    ovary_mask = ovary_image.astype(np.bool)
    cropped_label, reference_shape, offset = crop_image(
        label_image, ovary_mask, min_shape=min_shape
    )

    return cropped_label, reference_shape, offset


def pad_im_to_min_size(
    im: np.ndarray, min_size: Union[List[int], np.ndarray] = [80, 80, 80]
) -> np.ndarray:

    min_size = np.asarray(min_size)

    size_diff = np.array(im.shape) - min_size

    if np.all(size_diff >= 0):
        return im
    else:
        # we take negative size_diff because we subtracted min_size from im.shape
        pad_amount = np.clip(-size_diff, a_min=0, a_max=None)
        pad_width = [
            (0, pad_amount[0]),
            (0, pad_amount[1]),
            (0, pad_amount[2]),
        ]
        return np.pad(
            im, pad_width=pad_width, mode='constant', constant_values=0
        )



import glob
import os
from typing import Tuple, Union

import pandas as pd
from skimage.util import img_as_ubyte
from tqdm.auto import tqdm

from ..io.hdf5_io import write_raw
from ..io.raw_im import load_raw_im
from ..utils.file_names import parse_z_slice
from ..utils.image import rescale_raw_im
from ..utils.time import date_to_datestring


def convert_raw_dataset(
    row: pd.Series,
    output_dir: str,
    target_px_size: Tuple[float, float, float],
    im_path_key: str = 'path_im',
    n_z_key: str = 'n_z_im',
) -> dict:
    patient_id = row['patient']
    cycle_id = row['cycle']
    side = row['side']
    day = row['day']
    raw_data_path = row[im_path_key]
    n_z = row[n_z_key]
    im_px_size = (row['px_size_z'], row['px_size_x'], row['px_size_y'])

    image_file_pattern = os.path.join(raw_data_path, '*.tif*')
    image_files = glob.glob(image_file_pattern)

    z_slices = []
    good_im_paths = []
    for im_slice in image_files:
        success, z_idx = parse_z_slice(im_slice)
        if success is True:
            z_slices.append(z_idx)
            good_im_paths.append(im_slice)
    assert len(z_slices) == n_z
    image_files_sorted = [f for _, f in sorted(zip(z_slices, good_im_paths))]
    raw_im = load_raw_im(image_files_sorted)

    # rescale image
    rescaled_im = img_as_ubyte(
        rescale_raw_im(raw_im, im_px_size, target_px_size)
    )

    date_string = date_to_datestring(day)
    filename_base = f'{patient_id}_{cycle_id}_{side}_{date_string}_raw.h5'
    file_path = os.path.join(output_dir, filename_base)

    write_raw(
        raw_im=raw_im,
        raw_px_size=im_px_size,
        raw_path=raw_data_path,
        rescaled_im=rescaled_im,
        rescaled_px_size=target_px_size,
        fname=file_path,
        raw_name='raw',
        rescaled_name='raw_rescaled',
        compression='gzip',
    )

    success = True

    result = {
        'patient': patient_id,
        'cycle': cycle_id,
        'side': side,
        'day': day,
        'path': raw_data_path,
        'converted_path': file_path,
        'rescaled_px_size_x': target_px_size[1],
        'rescaled_px_size_y': target_px_size[2],
        'rescaled_px_size_z': target_px_size[0],
        'success': success,
    }
    return result


def batch_convert_raw(
    dataset_table: pd.DataFrame,
    target_px_size: Tuple[float, float, float],
    output_dir: Union[str, os.PathLike],
) -> pd.DataFrame:
    """Convert a batch of raw images

    Parameters
    ----------
    dataset_table : pd.DataFrame
        The table of label images to convert. Must contain columns:
        'patient', 'cycle', 'side', 'day', 'path_labels'
    target_px_size : Tuple[float, float, float]
        The target voxel size in mm/px. Should be in ZYX order.
    output_dir : os.PathLike
        The path to the directory to save the converted files to.

    Returns
    -------
    dataset_table : pd.DataFrame
        The input dataset_table with a `converted_raw_im_path` added
        containing the path to the converted files.
    """

    file_paths = []
    for i, row in tqdm(dataset_table.iterrows(), total=len(dataset_table)):

        file_path = convert_raw_dataset(
            row=row, output_dir=output_dir, target_px_size=target_px_size
        )
        file_paths.append(file_path)
    dataset_table['converted_im_path'] = file_paths

    return dataset_table



import glob
import os
from typing import List, Union, Tuple

import h5py
import numpy as np
import pandas as pd
import umetrics
from scipy.special import softmax
from skimage.measure import regionprops_table

from ..io.hdf5_io import load_im_from_hdf
from .follicles import post_process_follicles, process_follicle_labels
from .ovary import post_process_ovary
from ..utils.submission_utils import get_fold_directories


def _calc_effective_diameter(volumes: np.ndarray) -> np.ndarray:
    """Calculate the effective diameter of an object from
    its volume.

    Parameters
    ----------
    volumes : np.ndarray
        The array of image volumes to calculate the diameters from.

    Returns
    -------
    diameters : np.ndarray
        The calculated diameters. These are index matched
        to volumes.
    """
    return 2 * np.power(volumes * 0.75 / np.pi, 1 / 3)


def _classify_diameters(diameters: np.ndarray) -> List[str]:
    # classify the follicles
    follicle_class = []
    for d in diameters:
        if d < 2:
            follicle_class.append('<2')
        elif (d >= 2) and (d < 5):
            follicle_class.append('2-5')
        elif (d >= 5) and (d <= 10):
            follicle_class.append('5-10')
        elif d > 10:
            follicle_class.append('>10')
        else:
            raise ValueError('bad follicle diameter')
    return follicle_class


def _classify_labels(
    label_im: np.ndarray,
    indices: Union[List[int], np.ndarray],
    voxel_size_mm3: float,
) -> pd.DataFrame:
    """Classify objects by their effective diameter. The classes
    are <2, 2-5, 5-10, and >10 mm.


    Parameters
    ----------
    label_im : np.ndarray
        The label image from which to calculate properties.
    indices : List[int]
        The label indices to classify.
    voxel_size_mm3 : float
        The volume of one voxel in the label_im in mm^3


    Returns
    -------
    follicle_class_table : pd.DataFrame
        The table containing the class for each follicle. The follicle has
        the following columns:
            label: the label value in the prediction image.
            volume_px: the volume of the object in voxels
            volume_mm3: the volume of the object in mm^3
            diameter_mm: the calculated effective diameter in mm
            follicle_class: the class of the follicle (<2, 2-5, 5-10, >10)
    """
    rp_table = regionprops_table(label_im, properties=('label', 'area'))

    areas_to_measure = []
    indices = [i for i in indices if i != 0]
    labels_list = list(rp_table['label'])
    for label_list_index in indices:
        label_index = labels_list.index(label_list_index)
        areas_to_measure.append(rp_table['area'][label_index])
    areas_to_measure = np.asarray(areas_to_measure)

    # calculate the follicle volume/diameter
    areas_mm = voxel_size_mm3 * areas_to_measure
    diameters = _calc_effective_diameter(areas_mm)
    follicle_class = _classify_diameters(diameters)

    follicle_class_table = pd.DataFrame(
        {
            'label': indices,
            'volume_px': areas_to_measure,
            'volume_mm3': areas_mm,
            'diameter_mm': diameters,
            'follicle_class': follicle_class,
        }
    )
    return follicle_class_table


def quantify_performance(
    ref_im: np.ndarray,
    pred_im: np.ndarray,
    im_path: str,
    voxel_size_mm3: float,
    iou_threshold: float = 0.5,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Measure the precision and recall of an image segmentation.

    Parameters
    ----------
    ref_im : np.ndarray
        The ground truth reference image. Must be a label image.
    pred_im: np.ndarray
        The predicted segmentation. Must be a label image.
    im_path : str
        The file path of the prediction image.
    voxel_size_mm3 : float
        The volume of a voxel im mm^3.
    iou_threshold : float
        The minimum intersection over union value for a candidate match
        to be considered a true positive. Default value is 0.5.

    Returns
    -------
    metrics_table : pd.DataFrame
        The table of classification for each follicle. This table
        has columns:
            label: the label value in the prediction image
            volume_px: the volume of the object in voxels
            volume_mm3: the volume of the object in mm^3
            diameter_mm: the calculated effective diameter in mm
            follicle_class: the class of the follicle (<2, 2-5, 5-10, >10)
            file_path: the path to the prediction file
            iou_thresh: the IOU threshold used for calling true positives
    true_positive_table : pd.DataFrame
        The table of true positive label pairs (ref and prediction).
        The table has columns:
            label_ref: the label value in the reference image
            label_pred: the label value in the predicted image
            file_path: the path to the prediction file
            iou_thresh: the IOU threshold used for calling true positives
    """
    result = umetrics.calculate(
        ref_im, pred_im, strict=True, iou_threshold=iou_threshold
    )

    # get the true positive labels for the pred im
    tp_labels = [r[1] for r in result.true_positives]

    tp_table = _classify_labels(
        label_im=pred_im, indices=tp_labels, voxel_size_mm3=voxel_size_mm3
    )
    tp_table['seg_class'] = 'tp'

    # get the false positive data
    fp_labels = result.false_positives
    fp_table = _classify_labels(
        label_im=pred_im, indices=fp_labels, voxel_size_mm3=voxel_size_mm3
    )
    fp_table['seg_class'] = 'fp'

    # get the false negative data
    fn_labels = result.false_negatives
    fn_table = _classify_labels(
        label_im=ref_im, indices=fn_labels, voxel_size_mm3=voxel_size_mm3
    )
    fn_table['seg_class'] = 'fn'

    # get the split follicles
    split_ref_indices = []
    split_pred_indices = []
    for ref_index, pred_indices in result.split_errors:
        split_ref_indices.append(ref_index)
        split_pred_indices += pred_indices
    split_ref_table = _classify_labels(
        label_im=ref_im,
        indices=split_ref_indices,
        voxel_size_mm3=voxel_size_mm3,
    )
    split_ref_table['seg_class'] = 'split_ref'
    split_pred_table = _classify_labels(
        label_im=pred_im,
        indices=split_pred_indices,
        voxel_size_mm3=voxel_size_mm3,
    )
    split_pred_table['seg_class'] = 'split_pred'

    # get the merged follicles
    merge_ref_indices = []
    merge_pred_indices = []
    for pred_index, ref_indices in result.merge_errors:
        merge_ref_indices += ref_indices
        merge_pred_indices.append(pred_index)
    merge_ref_table = _classify_labels(
        label_im=ref_im,
        indices=merge_ref_indices,
        voxel_size_mm3=voxel_size_mm3,
    )
    merge_ref_table['seg_class'] = 'merge_ref'
    merge_pred_table = _classify_labels(
        label_im=pred_im,
        indices=merge_pred_indices,
        voxel_size_mm3=voxel_size_mm3,
    )
    merge_pred_table['seg_class'] = 'merge_pred'

    # get the ref labels not matching
    # all_ref_labels = np.unique(ref_im)
    # fn_labels = result.false_negatives
    # tp_ref_labels = [r[0] for r in result.true_positives]
    # un_ref_labels = [
    #     l for l in all_ref_labels if not (l in tp_ref_labels) and not (l in fn_labels)
    # ]
    # un_ref_table = _classify_labels(
    #     label_im=ref_im,
    #     indices=un_ref_labels,
    #     voxel_size_mm3=voxel_size_mm3
    #
    # )
    # un_ref_table['seg_class'] = 'unassigned_ref'
    #
    # # get the pred labels not matching
    all_pred_labels = np.unique(pred_im)
    un_pred_labels = [
        label_index
        for label_index in all_pred_labels
        if not (label_index in tp_labels)
        and not (label_index in fp_labels)
        and not (label_index in merge_pred_indices)
    ]
    un_pred_table = _classify_labels(
        label_im=pred_im, indices=un_pred_labels, voxel_size_mm3=voxel_size_mm3
    )
    un_pred_table['seg_class'] = 'unassigned_pred'

    # make sure all labels accounted for
    ref_labels = set(np.unique(ref_im))
    tp_ref_labels = [r[0] for r in result.true_positives]
    found_ref_labels = set.union(
        set(tp_ref_labels),
        set(fn_labels),
        set(split_ref_indices),
        set(merge_ref_indices),
        {0},
    )
    try:
        assert ref_labels == found_ref_labels
    except AssertionError:
        print(f'ref: {ref_labels}\nfound ref: {found_ref_labels}')

    pred_labels = set(np.unique(pred_im))
    found_pred_labels = set.union(
        set(tp_labels),
        set(fp_labels),
        set(split_pred_indices),
        set(merge_pred_indices),
        set(un_pred_labels),
        {0},
    )
    try:
        assert pred_labels == found_pred_labels
    except AssertionError:
        print(f'pred: {pred_labels}\nfound pred: {found_pred_labels}')

    # concatenate into single table
    metrics_table = pd.concat(
        [
            tp_table,
            fp_table,
            fn_table,
            split_ref_table,
            split_pred_table,
            merge_ref_table,
            merge_pred_table,
            un_pred_table,
        ]
    )
    metrics_table['file_path'] = im_path
    metrics_table['iou_thresh'] = iou_threshold

    # create the table of true positives
    true_positive_table = pd.DataFrame(
        {
            'label_ref': tp_ref_labels,
            'label_pred': tp_labels,
        }
    )
    true_positive_table['file_path'] = im_path
    true_positive_table['iou_thresh'] = iou_threshold

    return metrics_table, true_positive_table


def calculate_segmentation_performance(
    metrics_table: pd.DataFrame,
) -> Tuple[float, float]:
    """Calculate precision and recall for a set of predictions.

    Parameters
    ----------
    metrics_table : pd.DataFrame
        The table of tp/fp results from a set of segmentations

    Returns
    -------
    recall : float
        The recall, TP / (TP + FN)
    precision : float
        The precision, TP / (TP + FP)
    """
    value_counts = metrics_table['seg_class'].value_counts()

    try:
        n_tp = value_counts['tp']
    except KeyError:
        n_tp = 0

    try:
        n_fn = value_counts['fn']
    except KeyError:
        n_fn = 0

    try:
        n_fp = value_counts['fp']
    except KeyError:
        n_fp = 0

    try:
        n_unassigned_pred = value_counts['unassigned_pred']
    except KeyError:
        n_unassigned_pred = 0

    try:
        n_unassigned_ref = value_counts['unassigned_ref']
    except KeyError:
        n_unassigned_ref = 0

    try:
        recall = n_tp / (n_tp + n_fn + n_unassigned_ref)
        precision = n_tp / (n_tp + n_fp + n_unassigned_pred)
    except ZeroDivisionError:
        recall = np.nan
        precision = np.nan

    return recall, precision


def calculate_segmentation_performance_by_class(
    metrics_table: pd.DataFrame,
) -> pd.DataFrame:
    """Calculate the segmentation performance for each follicle class
    in the data set.

    The supported follicle classes are <2, 2-5, 5-10, and >10 mm diameters.

    Parameters
    ----------
    metrics_table : pd.DataFrame
        The table of segmentation results for the dataset. Must have a 'follicle_class'
        column and a seg_class column.

    Returns
    -------
    seg_performance_table : pd.DataFrame
        The resulting prediction and recall for each class.
    """
    small_fol_metrics = metrics_table.loc[
        metrics_table['follicle_class'] == '<2'
    ]
    small_r, small_p = calculate_segmentation_performance(small_fol_metrics)

    small_ant_fol_metrics = metrics_table.loc[
        metrics_table['follicle_class'] == '2-5'
    ]
    small_ant_r, small_ant_p = calculate_segmentation_performance(
        small_ant_fol_metrics
    )

    med_fol_metrics = metrics_table.loc[
        metrics_table['follicle_class'] == '5-10'
    ]
    med_r, med_p = calculate_segmentation_performance(med_fol_metrics)

    large_fol_metrics = metrics_table.loc[
        metrics_table['follicle_class'] == '>10'
    ]
    large_r, large_p = calculate_segmentation_performance(large_fol_metrics)

    seg_performance_table = pd.DataFrame(
        {
            'recall_<2': [small_r],
            'precision_<2': [small_p],
            'recall_2-5': [small_ant_r],
            'precision_2-5': [small_ant_p],
            'recall_5-10': [med_r],
            'precision_5-10': [med_p],
            'recall_>10': [large_r],
            'precision_>10': [large_p],
        }
    )
    return seg_performance_table


def quantify_performance_from_folder(
    prob_thresholds: List[float],
    iou_thresh: float,
    volume_threshold: float,
    voxel_size_mm3: float,
    predictions_dir: str,
    prediction_index: int,
    ref_dir: str,
    ref_file_pattern: str = '*.h5',
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Quantify segmentation performance

    Parameters
    ----------
    prob_thresholds
    iou_thresh
    volume_threshold
    voxel_size_mm3
    predictions_dir
    prediction_index
    ref_dir
    ref_file_pattern

    Returns
    -------

    """
    all_ref_fnames = glob.glob(os.path.join(ref_dir, ref_file_pattern))

    all_metrics_data = []
    all_true_positives = []
    for raw_fpath in all_ref_fnames:
        prediction_fname = os.path.basename(raw_fpath).replace(
            '.h5', '_predictions.h5'
        )
        prediction_fpath = os.path.join(predictions_dir, prediction_fname)
        with h5py.File(prediction_fpath, 'r') as f_pred:
            prediction_im = f_pred['predictions'][:]

        with h5py.File(raw_fpath, 'r') as f_raw:
            follicle_labels = f_raw['follicle_labels_rescaled'][:]
            ovary_labels = f_raw['ovary_labels_rescaled'][:]

        follicle_labels_clean = process_follicle_labels(
            follicle_labels=follicle_labels, volume_threshold=volume_threshold
        )

        metrics = []
        true_positives = []
        for thresh in prob_thresholds:
            follicle_seg = post_process_follicles(
                follicle_prediction=prediction_im,
                ovary_seg=ovary_labels,
                prob_threshold=thresh,
                volume_threshold=volume_threshold,
                prediction_index=prediction_index,
            )

            metrics_table, tp_table = quantify_performance(
                ref_im=follicle_labels_clean,
                pred_im=follicle_seg,
                im_path=prediction_fpath,
                voxel_size_mm3=voxel_size_mm3,
                iou_threshold=iou_thresh,
            )
            metrics_table['prob_threshold'] = thresh
            metrics.append(metrics_table)

            tp_table['prob_threshold'] = thresh
            true_positives.append(tp_table)

        image_metrics = pd.concat(metrics)
        all_metrics_data.append(image_metrics)

        all_true_positives.append(pd.concat(true_positives))

    all_metrics = pd.concat(all_metrics_data)
    all_tp_table = pd.concat(all_true_positives)

    all_seg_metrics = []

    for thresh in prob_thresholds:
        thresh_metrics = all_metrics.loc[
            all_metrics['prob_threshold'] == thresh
        ]

        seg_metrics = calculate_segmentation_performance_by_class(
            thresh_metrics
        )
        seg_metrics['prob_threshold'] = thresh

        all_seg_metrics.append(seg_metrics)

    seg_metrics = pd.concat(all_seg_metrics)

    return all_metrics, all_tp_table, seg_metrics


def quantify_performance_from_crossval(
    crossval_dir: str,
    fold_dir_pattern: str,
    prob_thresholds: List[float],
    iou_thresh: float,
    volume_threshold: float,
    voxel_size_mm3: float,
    predictions_dir_name: str,
    prediction_index: int,
    ref_dir_name: str,
    ref_file_pattern: str = '*.h5',
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    fold_directories = get_fold_directories(
        base_dir=crossval_dir, fold_dir_pattern=fold_dir_pattern
    )
    performance_tables = []
    tp_tables = []
    seg_metrics_tables = []

    for fold_index, fold_directory in fold_directories:
        ref_dir_path = os.path.join(fold_directory, ref_dir_name)
        prediction_dir_path = os.path.join(
            fold_directory, predictions_dir_name
        )

        (
            fold_performance_metrics,
            fold_tp_table,
            fold_seg_metrics,
        ) = quantify_performance_from_folder(
            prob_thresholds=prob_thresholds,
            iou_thresh=iou_thresh,
            volume_threshold=volume_threshold,
            voxel_size_mm3=voxel_size_mm3,
            predictions_dir=prediction_dir_path,
            prediction_index=prediction_index,
            ref_dir=ref_dir_path,
            ref_file_pattern=ref_file_pattern,
        )

        # add the fold index and path
        fold_performance_metrics['fold_index'] = fold_index
        fold_performance_metrics['fold_path'] = fold_directory
        fold_tp_table['fold_index'] = fold_index
        fold_tp_table['fold_path'] = fold_directory
        fold_seg_metrics['fold_index'] = fold_index
        fold_seg_metrics['fold_path'] = fold_directory

        performance_tables.append(fold_performance_metrics)
        tp_tables.append(fold_tp_table)
        seg_metrics_tables.append(fold_seg_metrics)

    return (
        pd.concat(performance_tables, ignore_index=True),
        pd.concat(tp_tables, ignore_index=True),
        pd.concat(seg_metrics_tables, ignore_index=True),
    )


def measure_iou(
    mask_reference: np.ndarray, mask_prediction: np.ndarray
) -> float:
    """Calculate the intersection over union of a segmentation.

    Parameters
    ----------
    mask_reference : np.ndarray
        The reference segmentation.
    mask_prediction : np.ndarray
        The predicted segmentation.

    Returns
    -------
    iou : float
        The intersection over union.
    """
    intersection = np.logical_and(mask_reference, mask_prediction)
    union = np.logical_or(mask_reference, mask_prediction)
    return np.sum(intersection) / np.sum(union)


def measure_iou_from_fold(
    fold_dir_path: str,
    ref_dir: str,
    prediction_dir: str,
    thresholds: List[float],
    prediction_index: int = 1,
    apply_softmax: bool = True,
    label_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
) -> pd.DataFrame:
    # get the files in the raw dir
    ref_dir_path = os.path.join(fold_dir_path, ref_dir)
    ref_file_pattern = os.path.join(ref_dir_path, "*.h5")
    ref_files = glob.glob(ref_file_pattern)

    # iterate through the raw dir
    ious = []
    table_thresholds = []
    file_paths = []
    fold_dirs = []
    for ref_fpath in ref_files:
        # load the raw image
        ref_im = load_im_from_hdf(ref_fpath, label_key)

        # load the prediction
        prediction_fname = os.path.basename(ref_fpath).replace(
            ".h5", "_predictions.h5"
        )
        prediction_fpath = os.path.join(
            fold_dir_path, prediction_dir, prediction_fname
        )
        prediction_im = load_im_from_hdf(prediction_fpath, prediction_key)

        if apply_softmax is True:
            prediction_im = softmax(prediction_im, axis=0)
        prediction_im = prediction_im[prediction_index, ...]

        for thresh in thresholds:
            # segment the prediction
            prediction_mask = post_process_ovary(
                ovary_prediction=prediction_im,
                raw_im=None,
                threshold=thresh,
                dilation_size=None,
                mask_padding=False,
            )
            ref_mask = ref_im.astype(bool)

            ious.append(measure_iou(ref_mask, prediction_mask))
            table_thresholds.append(thresh)
            file_paths.append(prediction_fpath)
            fold_dirs.append(fold_dir_path)

    return pd.DataFrame(
        {
            'fold_path': fold_dirs,
            'file_path': file_paths,
            'threshold': table_thresholds,
            'iou': ious,
        }
    )


def measure_iou_from_crossval(
    crossval_dir: str,
    fold_dir_pattern: str,
    ref_dir: str,
    prediction_dir: str,
    thresholds: List[float],
    prediction_index: int = 1,
    apply_softmax: bool = True,
    label_key: str = 'ovary_labels_rescaled',
    prediction_key: str = 'predictions',
) -> pd.DataFrame:
    fold_directories = get_fold_directories(
        base_dir=crossval_dir, fold_dir_pattern=fold_dir_pattern
    )

    measurement_tables = []
    for fold_index, fold_directory in fold_directories:
        measurement_tables.append(
            measure_iou_from_fold(
                fold_dir_path=fold_directory,
                ref_dir=ref_dir,
                prediction_dir=prediction_dir,
                thresholds=thresholds,
                prediction_index=prediction_index,
                apply_softmax=apply_softmax,
                label_key=label_key,
                prediction_key=prediction_key,
            )
        )
    return pd.concat(measurement_tables)



import numpy as np
import pytest

from ovary_analysis.utils.image import pad_label_im, pad_im_to_min_size


def test_pad_label_im():
    raw_im = np.random.random((10, 20, 20))
    label_im = np.random.random((8, 20, 20)).astype(int)

    label_im_2 = pad_label_im(raw_im, label_im)
    assert label_im_2.shape == raw_im.shape


def test_pad_label_im_no_pad():
    raw_im_shape = (10, 20, 20)
    raw_im = np.random.random(raw_im_shape)
    label_im = np.random.random(raw_im_shape).astype(int)

    label_im_2 = pad_label_im(raw_im, label_im)
    assert label_im_2.shape == raw_im.shape
    assert label_im_2 is label_im


def test_pad_label_im_bad_shape():
    raw_im = np.random.random((10, 20, 20))
    label_im = np.random.random((20, 20, 20)).astype(int)

    with pytest.raises(AssertionError):
        _ = pad_label_im(raw_im, label_im)


def test_pad_image_min_size_no_pad():
    test_im = np.random.random((100, 80, 90))
    min_size = [80, 80, 80]

    padded_im = pad_im_to_min_size(test_im, min_size)

    assert padded_im is test_im
    np.testing.assert_array_equal(padded_im.shape, test_im.shape)


def test_pad_image():
    test_im = np.random.random((100, 60, 90))
    min_size = [80, 80, 80]

    padded_im = pad_im_to_min_size(test_im, min_size)

    np.testing.assert_array_equal(padded_im.shape, [100, 80, 90])



import os
from typing import List, Optional, Tuple
import numpy as np
from skimage import io
import datetime
import warnings


def _validate_patient_id(patient_id: str) -> str:
    """check if patient id is in right format A{number} or B{number}

    Parameters
    ----------
    patient_id: str
        Patient ID

    Returns
    -------
    patient_id: str
        return a str if the format is right, None if the format wrong.
    """
    if (
        patient_id[0] == 'A'
        or patient_id[0] == 'B'
        and patient_id[1:].isdecimal()
    ):
        return patient_id
    else:
        warnings.warn("patient ID format error")
        return None


def _validate_cycle(cycle: str) -> str:
    """check if cycle is in right format z1 or z2

    Parameters
    ----------
    cycle: str
        z1 or z2

    Returns
    -------
    cycle: str
        return a str if the format is right, None if the format wrong.
    """
    if cycle == 'z1' or cycle == 'z2':
        return cycle
    else:
        warnings.warn("cycle format error")
        return None


def _validate_side(side: str) -> str:
    """check if side is in right format links or rechts

    Parameters
    ----------
    side: str
        links or rechts

    Returns
    -------
    side: str
        return a str if the format is right, None if the format wrong.
    """
    if side == 'rechts' or side == 'links':
        return side
    else:
        warnings.warn("side format error")
        return None


def _validate_date(date: str) -> datetime.date:
    """check if date is in right format YYMMDD and transform it in datatime

    Parameters
    ----------
    date: str
        YYMMDD

    Returns
    -------
    date: datetime.date
        return a datetime.date if the format is right, None if the format wrong.
    """
    try:
        date = datetime.datetime.strptime(date, "%y%m%d").date()
        return date
    except Exception as e:
        warnings.warn(str(e))
        return None


def parse_seg_fname(fname: str) -> Tuple[str, str, str, datetime.date]:
    """parse the segmentation of file name

    Parameters
    ----------
    fname: str
        file name in form of <patient_id> _ <cycle> _ <side> _ <date>_raw_follicles.h5

    Returns
    -------
    patientID: str
        ID of patient
    cycle: str
        z1 or z2 cycle
    side: str
        links or rechts (left or right) ovary
    date: datetime.date
        the date when the image is taken in form of YYMMDD
    """

    patient_id = _validate_patient_id(fname.split('_')[0])

    cycle = _validate_cycle(fname.split('_')[1].lower())

    side = _validate_side(fname.split('_')[2].lower())

    date = _validate_date(fname.split('_')[3].lower())

    return patient_id, cycle, side, date


def parse_slice_from_raw_fname(fname: str) -> int:
    f_base_name = os.path.basename(fname).lower()
    z_slice_with_ext = f_base_name.split('_z')[-1]
    z_slice = os.path.splitext(z_slice_with_ext)[0]

    return int(z_slice)


def parse_data_directory_name(path: str) -> Tuple[str, str, str]:
    assert os.path.isdir(path)

    norm_path = os.path.normpath(path)
    folders = norm_path.split(os.sep)

    date = folders[-1].replace('.', '')

    ovary_data = folders[-2].split(' ')
    cycle = ovary_data[0].lower()
    side = ovary_data[1].lower()

    return date, cycle, side


def is_cycle_folder(folder_path: str) -> bool:
    """Verify the directory is a valid cycle folder"""
    folder_name = os.path.basename(folder_path).lower()
    contains_cycle = ('z1' in folder_name) != ('z2' in folder_name)
    contains_side = ('links' in folder_name) != ('rechts' in folder_name)

    if contains_cycle and contains_side:
        return True
    else:
        return False


def parse_cycle_folder(folder_path: str) -> Tuple[str, str]:
    if is_cycle_folder(folder_path):
        folder_name = os.path.basename(folder_path).lower()

        if 'z1' in folder_name:
            cycle = 'z1'
        else:
            cycle = 'z2'
        if 'links' in folder_name:
            side = 'links'
        else:
            side = 'rechts'
    else:
        raise ValueError()

    return cycle, side


def parse_day_dirs(
    day_dirs: List[str],
) -> Tuple[List[datetime.date], List[str]]:
    date_list = []
    good_day_dirs = []
    bad_day_dirs = []
    for d in day_dirs:
        dir_name = os.path.basename(d)
        success, formatted_date = _parse_day(dir_name)
        if success is True:
            date_list.append(formatted_date)
            good_day_dirs.append(d)
        else:
            bad_day_dirs.append(d)
    return date_list, good_day_dirs, bad_day_dirs


def _parse_day(day_dir_name: str) -> Tuple[bool, Optional[datetime.date]]:
    try:
        name_parts = day_dir_name.split('.')
        day = int(name_parts[0])
        month = int(name_parts[1])
        year = 2000 + int(name_parts[2])
        formatted_date = datetime.date(day=day, month=month, year=year)
        success = True
    except (ValueError, TypeError):
        success = False
        formatted_date = None

    return success, formatted_date


def calculate_day_offsets(date_list: List[datetime.date]) -> List[int]:
    first_date = min(date_list)
    offset_list = [
        _calculate_day_offset(d, first_date=first_date) for d in date_list
    ]

    return offset_list


def _calculate_day_offset(
    date: datetime.date, first_date: datetime.date
) -> int:
    date_delta = date - first_date
    offset = date_delta.days
    return offset


def parse_z_slice(fname: str) -> int:
    try:
        f_base_name = os.path.basename(fname).lower()
        z_slice_with_ext = f_base_name.split('_z')[-1]
        z_slice = os.path.splitext(z_slice_with_ext)[0]
        slice_index = int(z_slice)
        success = True
    except (ValueError, IndexError):
        success = False
        slice_index = None

    return success, slice_index


def validate_z_slices(z_slices: List[int]) -> bool:
    sorted_slices = sorted(z_slices)
    z_diff = np.diff(sorted_slices)

    # the z slices should be contiguous and start at 0
    if len(sorted_slices) > 0:
        is_valid = np.all(z_diff == 1) and (sorted_slices[0] == 0)
    else:
        is_valid = False

    return is_valid


def parse_label_file_name(fname: str) -> datetime.date:
    f_base = os.path.basename(fname)
    date_parts = f_base.split(' ')[0].split('.')

    day = int(date_parts[0])
    month = int(date_parts[1])
    year = 2000 + int(date_parts[2])

    return datetime.date(day=day, month=month, year=year)


def validate_labels_file(im_path: str) -> Tuple[bool, int]:
    im = io.imread(im_path)
    im_shape = im.shape

    # there should be two labels channels
    valid_im = im_shape[1] == 2
    n_z = im_shape[0]

    return valid_im, n_z



import datetime
from typing import Union


def date_to_datestring(day: Union[str, datetime.date]) -> str:
    """Convert a date to a %y%m%d formatted string.

    Parameters
    ----------
    day : Union[str, datetime.date]
        The date to convert. If day is a string, it should be
        formatted '%Y-%m-%d'

    Returns
    -------
    datestring : str
        The date formatted %y%m%d. For example November 9, 2020 would
        be formatted 201109.
    """
    if isinstance(day, str):
        date_time_obj = datetime.datetime.strptime(day, '%Y-%m-%d')
        date_string = date_time_obj.strftime('%y%m%d')
    else:
        date_string = day.strftime('%y%m%d')
    return date_string



import os
from shutil import copy2
from typing import List, Union

import numpy as np
import pandas as pd


def split_table_into_k_folds(
    df: pd.DataFrame, n_folds: int = 10, random_seed: int = 42
) -> List[pd.DataFrame]:
    """Split a table of datasets into k folds

    Parameters
    ----------
    df : pd.DataFrame
        The table to split.
    n_folds : int
        The number of folds to split the dataframe into.
    random_seed : int
        The seed for the random number generator used to shuffle the datasets.
        The default value is 42.

    Returns
    -------
    split_table : List[pd.DataFrame]
        Each fold of the split dataframe stored in a list.
    """
    shuffled_df = df.sample(frac=1, random_state=random_seed)
    return np.array_split(shuffled_df, n_folds)


def copy_datasets(
    df: pd.DataFrame,
    in_dir: Union[str, os.PathLike],
    out_dir: Union[str, os.PathLike],
    column_name: str = 'dataset_path',
):
    if not os.path.isdir(out_dir):
        os.mkdir(out_dir)
    for _, row in df.iterrows():
        dataset_name = os.path.basename(row[column_name])
        from_path = os.path.join(in_dir, dataset_name)
        to_path = os.path.join(out_dir, dataset_name)
        copy2(from_path, to_path)



import os

from ovary_analysis.post_process.segmentation_metrics import quantify_performance_from_crossval


CROSSVAL_DIR = '/cluster/work/cobi/kevin/ivf/cross_validation_20220126/follicle_crossval_no_mask'
FOLD_DIR_PATTERN = 'fold_*'
PROB_THRESHOLDS = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
IOU_THRESH = 0.5
VOLUME_THRESHOLD = 30
VOXEL_SIZE_MM3 = 0.157288 * 0.157288 * 0.157288
PREDICTIONS_DIR_NAME = 'predictions'
PREDICTIONS_INDEX = 1
REF_DIR_NAME = 'test'
REF_FILE_PATTERN = '*.h5'
OUTPUT_NAME_BASE = 'follicle_crossval_no_mask'

performance_metrics, tp_table, seg_metrics = quantify_performance_from_crossval(
        crossval_dir=CROSSVAL_DIR,
        fold_dir_pattern=FOLD_DIR_PATTERN,
        prob_thresholds=PROB_THRESHOLDS,
        iou_thresh=IOU_THRESH,
        volume_threshold=VOLUME_THRESHOLD,
        voxel_size_mm3=VOXEL_SIZE_MM3,
        predictions_dir_name=PREDICTIONS_DIR_NAME,
        prediction_index=PREDICTIONS_INDEX,
        ref_dir_name=REF_DIR_NAME,
        ref_file_pattern = REF_FILE_PATTERN,
)

output_name_performance_metrics = OUTPUT_NAME_BASE + '_performance_metrics.csv'
output_path_performance_metrics = os.path.join(CROSSVAL_DIR, output_name_performance_metrics)
performance_metrics.to_csv(output_path_performance_metrics)

output_name_tp_table = OUTPUT_NAME_BASE + '_tp_table.csv'
output_path_tp_table = os.path.join(CROSSVAL_DIR, output_name_tp_table)
tp_table.to_csv(output_path_tp_table)

output_name_seg_metrics = OUTPUT_NAME_BASE + '_seg_metrics.csv'
output_path_seg_metrics = os.path.join(CROSSVAL_DIR, output_name_seg_metrics)
seg_metrics.to_csv(output_path_seg_metrics)



import pandas as pd
import numpy as np
from ovary_analysis.utils.file_names import calculate_day_offsets


def date_csv_to_df(csv_dir: str) -> pd.DataFrame:
    """import csv file to pandas dataframe
    and assign datetime dtype to column date

    Parameters
    ----------
    csv_dir: str
        directory of the csv file

    Returns
    -------
    raw_data: pandas dataframe
        dataframe with follicle info from each images

    """
    raw_data = pd.read_csv(csv_dir)
    raw_data['date'] = pd.to_datetime(raw_data['date'])
    return raw_data


def calc_relative_day(df: pd.DataFrame) -> pd.DataFrame:
    """calculate the relative days of each patients in each cycles

    Parameters
    ----------
    df: pandas Dataframe
        dataframe with follicle info from each images

    Returns
    -------
    df: pandas Dataframe
        add relative days info to the original dataframe

    """
    df['day_rel'] = np.nan
    z1 = df.loc[df['cycle'] == 'z1']
    z2 = df.loc[df['cycle'] == 'z2']
    z1_patient_id = z1['patient_id'].unique().tolist()
    z2_patient_id = z2['patient_id'].unique().tolist()

    for id1 in z1_patient_id:
        z1_links = (
            df['date']
            .loc[
                (df['cycle'] == 'z1')
                & (df['patient_id'] == id1)
                & (df['side'] == 'links')
            ]
            .tolist()
        )

        if len(z1_links) > 0:
            rel_days_links = calculate_day_offsets(z1_links)
            df['day_rel'].loc[
                (df['cycle'] == 'z1')
                & (df['side'] == 'links')
                & (df['patient_id'] == id1)
            ] = rel_days_links

        z1_rechts = (
            df['date']
            .loc[
                (df['cycle'] == 'z1')
                & (df['patient_id'] == id1)
                & (df['side'] == 'rechts')
            ]
            .tolist()
        )
        if len(z1_rechts) > 0:
            rel_days_rechts = calculate_day_offsets(z1_rechts)
            df['day_rel'].loc[
                (df['cycle'] == 'z1')
                & (df['side'] == 'rechts')
                & (df['patient_id'] == id1)
            ] = rel_days_rechts

    for id2 in z2_patient_id:
        z2_links = (
            df['date']
            .loc[
                (df['cycle'] == 'z2')
                & (df['patient_id'] == id2)
                & (df['side'] == 'links')
            ]
            .tolist()
        )

        if len(z2_links) > 0:
            rel_days_links = calculate_day_offsets(z2_links)
            df['day_rel'].loc[
                (df['cycle'] == 'z2')
                & (df['side'] == 'links')
                & (df['patient_id'] == id2)
            ] = rel_days_links

        z2_rechts = (
            df['date']
            .loc[
                (df['cycle'] == 'z2')
                & (df['patient_id'] == id2)
                & (df['side'] == 'rechts')
            ]
            .tolist()
        )

        if len(z2_rechts) > 0:
            rel_days_rechts = calculate_day_offsets(z2_rechts)
            df['day_rel'].loc[
                (df['cycle'] == 'z2')
                & (df['side'] == 'rechts')
                & (df['patient_id'] == id2)
            ] = rel_days_rechts

    return df


def process_fol_csv(csv_dir: str):
    """import csv file and calculate relative days

    Parameters
    ----------
    csv_dir: str
        directory of the csv file

    Returns
    -------
    df: pandas Dataframe
        add relative days info to the imported dataframe

    """
    df = date_csv_to_df(csv_dir)
    print(df.info())
    df = voxel_to_mm(df)
    df = calc_relative_day(df).sort_values(
        by=['patient_id', 'cycle', 'side', 'day_rel']
    )
    return df


def remove_dominant_follicle(df: pd.DataFrame) -> pd.DataFrame:
    """drop the dominant follicle of each patients on each relative day

    Parameters
    ----------
    df: pandas Dataframe
        dataframe with follicle info from each images

    Returns
    -------
    df: pandas Dataframe
        dataframe with follicle info from each images but exclude dominant follicles

    """
    cycles = df['cycle'].unique().tolist()
    day_rel = df['day_rel'].unique().tolist()
    patient_id = df['patient_id'].unique().tolist()

    for c in cycles:
        for day in day_rel:
            for patient in patient_id:
                try:
                    domi = df.loc[
                        (df['cycle'] == c)
                        & (df['day_rel'] == day)
                        & (df['patient_id'] == patient)
                    ]
                    ind = domi['volume'].idxmax()
                    df = df.drop(ind)
                except ValueError:
                    None
    return df.sort_values(
        by=['patient_id', 'cycle', 'side', 'day_rel'], ignore_index=True
    )


def unique_patient(
    df: pd.DataFrame, keep='last', level='day_rel'
) -> pd.DataFrame:
    """Determine which datasets each patient ID has. This function checks for days, sides, and cycles.

    Parameters
    ----------
    df: pandas dataframe
        follicle data include side, cycle, patient_id and relative day

    Returns
    -------
    df_unique: pandas dataframe
        follicle data in which no duplicate patient id on each relative day.

    """
    df = df.sort_values(
        by=['patient_id', 'cycle', 'side', 'day_rel'], ignore_index=True
    )
    if level == 'day_rel':
        df_unique = df.drop_duplicates(
            subset=['side', 'cycle', 'day_rel', 'patient_id'], keep=keep
        )
    elif level == 'cycle':
        df_unique = df.drop_duplicates(
            subset=['side', 'cycle', 'patient_id'], keep=keep
        )
    return df_unique.sort_values(
        by=['patient_id', 'cycle', 'side', 'day_rel'], ignore_index=True
    )


def extract_dominant(df: pd.DataFrame, level: str = 'day_rel') -> pd.DataFrame:
    """pick out the dominant follicle of each patients with respect to relative day or cycle

    Parameters
    ----------
    df: pandas Dataframe
        dataframe with follicle info from each images

    Returns
    -------
    dominant_df: pandas Dataframe
        dataframe with dominant follicle info of each patients with respect to relative day or cycle

    """
    dominant_df = pd.DataFrame()
    sides = ['links', 'rechts']
    cycles = df['cycle'].unique().tolist()
    day_rel = df['day_rel'].unique().tolist()
    patient_id = df['patient_id'].unique().tolist()
    if level == 'day_rel':
        for c in cycles:
            for day in day_rel:
                for patient in patient_id:
                    for side in sides:
                        try:
                            domi = df.loc[
                                (df['cycle'] == c)
                                & (df['day_rel'] == day)
                                & (df['patient_id'] == patient)
                                & (df['side'] == side)
                            ]
                            ind = domi['volume'].idxmax()
                            dominant_df = dominant_df.append(domi.loc[ind])
                        except ValueError:
                            None
    elif level == 'cycle':
        for c in cycles:
            for patient in patient_id:
                for side in sides:
                    try:
                        domi = df.loc[
                            (df['cycle'] == c)
                            & (df['patient_id'] == patient)
                            & (df['side'] == side)
                        ]
                        ind = domi['volume'].idxmax()
                        temp_domi = domi.loc[ind]
                        temp_domi['original_index'] = ind
                        dominant_df = dominant_df.append(temp_domi)
                    except ValueError:
                        None
    return dominant_df.sort_values(
        by=['patient_id', 'cycle', 'side', 'day_rel'], ignore_index=True
    )


def voxel_to_mm(df: pd.DataFrame) -> pd.DataFrame:
    """unit conversion from voxel to mm^3

    Parameters
    ----------
    df: pandas dataframe
        follicle data which record volume in voxel

    Returns
    -------
    df: pandas dataframe
        follicle data which record volume in mm^3

    """
    df['volume'] = df['volume'].map(lambda x: x * pow(0.157, 3))
    return df


def calc_growth(df: pd.DataFrame) -> pd.DataFrame:
    """calculate the growth of only dominant follicle from one side of ovary of one patient.
        input dataframe should includes only data from dominant follicles

    Parameters
    ----------
    df: pandas dataframe
        follicle data, information such as patient_id, relative days, cycle, side and volume is needed.

    Returns
    -------
    df: pandas dataframe
        add one column named growth to data frame which contains info about the growth of follicle size at each time points.

    """
    df = df.sort_values(by=['patient_id', 'cycle', 'side', 'day_rel'])
    sides = ['links', 'rechts']
    cycles = ['z1', 'z2']
    temp_df_growth = pd.DataFrame()
    for cycle in cycles:
        id_list = df['patient_id'].loc[df['cycle'] == cycle].unique().tolist()
        for patient in id_list:
            for side in sides:
                cycle_volume_df = df.loc[
                    (df['patient_id'] == patient)
                    & (df['cycle'] == cycle)
                    & (df['side'] == side)
                ]
                cycle_volume_df = cycle_volume_df.sort_values(
                    by=['day_rel'], ignore_index=True
                )
                cycle_volume_df['growth'] = np.nan
                if len(cycle_volume_df) > 0:
                    growth_array = np.array([0])
                    if len(cycle_volume_df) > 1:
                        growth_array = np.append(
                            growth_array, np.diff(cycle_volume_df['volume'])
                        )
                    cycle_volume_df['growth'].loc[
                        (cycle_volume_df['patient_id'] == patient)
                        & (cycle_volume_df['cycle'] == cycle)
                        & (cycle_volume_df['side'] == side)
                    ] = growth_array

                    temp_df_growth = temp_df_growth.append(
                        cycle_volume_df[
                            [
                                'patient_id',
                                'cycle',
                                'side',
                                'day_rel',
                                'growth',
                            ]
                        ]
                    )
    df = pd.merge(
        df, temp_df_growth, on=['patient_id', 'cycle', 'side', 'day_rel']
    )
    return df


def calc_domi_relative_day(df: pd.DataFrame) -> pd.DataFrame:
    """calculate the day relative to the day the largest dominant follicle occurred on
    for a given cycle and one side of ovary

    Parameters
    ----------
    df: pandas dataframe
        follicle data, information such as patient_id, relative days, cycle, side and volume is needed.

    Returns
    -------
    df: pandas dataframe
        add one column named domi_day_rel to data frame which contains info about
        the day relative to the day the largest dominant follicle occurred on.

    """
    df['day_from_peak'] = np.nan
    domi_cycle = extract_dominant(df, level='cycle')
    sides = ['links', 'rechts']
    cycles = df['cycle'].unique().tolist()
    patient_id = df['patient_id'].unique().tolist()
    for c in cycles:
        for patient in patient_id:
            for side in sides:
                try:
                    domi_cycle_day = (
                        domi_cycle['day_rel']
                        .loc[
                            (domi_cycle['cycle'] == c)
                            & (domi_cycle['patient_id'] == patient)
                            & (domi_cycle['side'] == side)
                        ]
                        .tolist()
                    )

                    df['day_from_peak'].loc[
                        (df['cycle'] == c)
                        & (df['patient_id'] == patient)
                        & (df['side'] == side)
                    ] = (
                        df['day_rel'].loc[
                            (df['cycle'] == c)
                            & (df['patient_id'] == patient)
                            & (df['side'] == side)
                        ]
                        - domi_cycle_day[0]
                    )
                except IndexError:
                    None
    return df


def group_parse(df: pd.DataFrame) -> pd.DataFrame:
    """parse patient_id adn create a column called growth comprised of "A" or "B"

    Parameters
    ----------
    df: pandas dataframe
        follicle data, information such as patient_id is needed

    Returns
    -------
    df: pandas dataframe
        orignal dataframe with new column called group

    """
    patient_id_list = df['patient_id'].tolist()
    group_list = [patient[0] for patient in patient_id_list]
    df['group'] = group_list
    return df



import glob
import os
from typing import Union, List, Dict, Any, Tuple

TRAIN_SCRIPT_TEMPLATE = """
#!/bin/bash

# load the python module
module load gcc/6.3.0 python_gpu/3.8.5

python {train_runner} --config {config_path}
"""

PREDICT_SCRIPT_TEMPLATE = """
#!/bin/bash

# load the python env
module load gcc/6.3.0 python_gpu/3.8.5

python {predict_runner} --config {config_path}
"""

SUBMISSION_FILE_STUB = """
#!/bin/bash

env2lmod

# load the python env
module load gcc/6.3.0 python_gpu/3.8.5

"""
SUBMISSION_LINE_TEMPLATE = (
    'bsub -n {n_cpus} -W {hours:02d}:{minutes:02d} -N -B '
    + '-J "{job_name}" -R "rusage[mem={mem_per_cpu}, ngpus_excl_p={n_gpus}]" '
    + '-R "select[gpu_model0==GeForceRTX2080Ti]" < {runner_path}\n'
)


def _write_predict_script(
    predict_runner: str,
    config_path: Union[str, os.PathLike],
    output_dir: Union[str, os.PathLike],
) -> str:
    """Write the bash script to run prediction

    Parameters
    ----------
    predict_runner : str
        The command to call the predict runner. This us generally a bash
        or python script.
    config_path : Union[str, os.PathLike]
        The path to the configuration file.
    output_dir : Union[str, os.PathLike]
        The path to folder the bash script will be saved.

    Returns
    -------
    script_path : str
        The path the bash script was saved to.
    """
    predict_script = PREDICT_SCRIPT_TEMPLATE.format(
        predict_runner=predict_runner, config_path=config_path
    )
    script_path = os.path.join(output_dir, 'run_prediction.sh')
    with open(script_path, 'w') as out_file:
        out_file.write(predict_script)

    return script_path


def _write_train_script(
    train_runner: str,
    config_path: Union[str, os.PathLike],
    output_dir: Union[str, os.PathLike],
) -> str:
    """Write the bash script to run prediction

    Parameters
    ----------
    train_runner : str
        The command to call the train runner. This us generally a bash
        or python script.
    config_path : Union[str, os.PathLike]
        The path to the configuration file.
    output_dir : Union[str, os.PathLike]
        The path to folder the bash script will be saved.

    Returns
    -------
    script_path : str
        The path the bash script was saved to.
    """
    train_script = TRAIN_SCRIPT_TEMPLATE.format(
        train_runner=train_runner, config_path=config_path
    )
    script_path = os.path.join(output_dir, 'run_training.sh')
    with open(script_path, 'w') as out_file:
        out_file.write(train_script)

    return script_path


def _write_submission_script(
    runner_paths: List[str],
    n_cpus: int,
    mem_per_cpu: int,
    n_gpus: int,
    job_time: List[int],
    job_prefix: str,
    script_path: Union[str, os.PathLike],
):
    """Write a script to submit a batch of jobs

    Parameters
    ----------
    runner_paths : List[str]
        The list of paths to the runner scripts to submit.
    n_cpus : int
        The number of CPUS to request for each job.
    mem_per_cpu : int
        The amount of RAM per CPU to request in MB.
    n_gpus : int
        The number of GPUS to request.
    job_time : List[int]
        The maximum job wall clock time in [hours, minutes].
    job_prefix : str
        The prefix used for naming the job. Job names will be
        {job_prefix}_i
    script_path : Union[str, os.PathLike]
        The path to save the script to.
    """
    submission_script = SUBMISSION_FILE_STUB

    for i, runner in enumerate(runner_paths):
        job_name = job_prefix + f'_{i}'
        submission_line = SUBMISSION_LINE_TEMPLATE.format(
            n_cpus=n_cpus,
            mem_per_cpu=mem_per_cpu,
            hours=job_time[0],
            minutes=job_time[1],
            n_gpus=n_gpus,
            job_name=job_name,
            runner_path=runner,
        )
        submission_script += submission_line

    with open(script_path, 'w') as out_file:
        out_file.write(submission_script)


def write_batch_submission_script(
    submission_script_path: str,
    runner_scripts: List[str],
    job_prefix: str,
    job_settings: Dict[str, Any],
):
    """Write a submission script for a set of jobs.

    Parameters
    ----------
    submission_script_path : str
        The path to which the submission script will be saved.
    runner_scripts : List[str]
        The list of runners to submit as jobs. One job will be created for
        each runner script.
    job_prefix : str
        The name to prepend to each submitted job.
    job_settings : Dict[str, Any]
        The LSF job settings for each job.
    """

    job_settings.update(
        {
            'runner_paths': runner_scripts,
            'job_prefix': job_prefix,
            'script_path': submission_script_path,
        }
    )
    _write_submission_script(**job_settings)


def _get_fold_index(fold_path: str, separator: str = '_') -> int:
    """Get the fold index from a fold directory (used for crossval).

    Parameters
    ----------
    fold_path : str
        The path to the folder
    separator : str
        The separator to split the folder name by. The last segment
        of the split string is returned as the fold index.
        The default value is '_'

    Returns
    -------
    fold_index : int
        The index for the fold contained in fold_path.
    """
    folder_name = os.path.basename(fold_path)
    return int(folder_name.split(separator)[-1])


def get_fold_directories(
    base_dir: str, fold_dir_pattern: str = 'fold_*'
) -> List[Tuple[int, str]]:
    """

    Parameters
    ----------
    base_dir : str
        The base directory containing the crossvalidation folds.
    fold_dir_pattern : str
        The pattern used to filter for fold directories. This is
        passed to glob. Default value is 'fold_*'

    Returns
    -------
    fold_dirs : List[Tuple[int, str]]
        The fold indices and folder directories. Each tuple is
        (fold_index, fold_dir).
    """
    folder_pattern = os.path.join(base_dir, fold_dir_pattern)
    fold_candidates = glob.glob(folder_pattern)

    # make sure they are directories
    fold_dirs = [
        (_get_fold_index(f), f) for f in fold_candidates if os.path.isdir(f)
    ]

    return fold_dirs
